"""
ubxvariants.py

Various routines to get payload dictionaries for message
types which exist in multiple variants for the same
message class, id and mode.

Created on 20 May 2024

:author: semuadmin
:copyright: SEMU Consulting © 2020
:license: BSD 3-Clause
"""

from pyubx2.exceptions import UBXMessageError
from pyubx2.ubxhelpers import val2bytes
from pyubx2.ubxtypes_core import GET, POLL, SET, U1, UBX_MSGIDS
from pyubx2.ubxtypes_get import UBX_PAYLOADS_GET
from pyubx2.ubxtypes_poll import UBX_PAYLOADS_POLL
from pyubx2.ubxtypes_set import UBX_PAYLOADS_SET


def get_cfgtp5_dict(**kwargs) -> dict:
    """
    Select appropriate CFG-TP5 POLL payload definition by checking
    presence of tpIdx or payload argument.

    :param kwargs: optional payload key/value pairs
    :return: dictionary representing payload definition
    :rtype: dict

    """

    lp = 0
    if "payload" in kwargs:
        lp = len(kwargs["payload"])
    elif "tpIdx" in kwargs:
        lp = 1
    print(f"DEBUG TP5 dict {kwargs} len payload {lp}")
    if lp == 1:
        return UBX_PAYLOADS_POLL["CFG-TP5-TPX"]
    return UBX_PAYLOADS_POLL["CFG-TP5"]  # pragma: no cover


def get_mga_dict(msg: bytes, mode: int, **kwargs) -> dict:
    """
    Select appropriate MGA payload definition by checking
    value of 'type' attribute (1st byte of payload).

    :param str mode: mode (0=GET, 1=SET, 2=POLL)
    :param kwargs: optional payload key/value pairs
    :return: dictionary representing payload definition
    :rtype: dict
    :raises: UBXMessageError

    """

    if "type" in kwargs:
        typ = val2bytes(kwargs["type"], U1)
    elif "payload" in kwargs:
        typ = kwargs["payload"][0:1]
    else:
        raise UBXMessageError(
            "MGA message definitions must include type or payload keyword"
        )
    identity = UBX_MSGIDS[msg + typ]
    if mode == SET:
        return UBX_PAYLOADS_SET[identity]
    return UBX_PAYLOADS_GET[identity]


def get_rxmpmreq_dict(**kwargs) -> dict:
    """
    Select appropriate RXM-PMREQ payload definition by checking
    the 'version' keyword or payload length.

    :param kwargs: optional payload key/value pairs
    :return: dictionary representing payload definition
    :rtype: dict
    :raises: UBXMessageError

    """

    lpd = 0
    if "version" in kwargs:  # assume longer version
        lpd = 16
    elif "payload" in kwargs:
        lpd = len(kwargs["payload"])
    else:
        raise UBXMessageError(
            "RXM-PMREQ message definitions must include version or payload keyword"
        )
    if lpd == 16:
        return UBX_PAYLOADS_SET["RXM-PMREQ"]  # long
    return UBX_PAYLOADS_SET["RXM-PMREQ-S"]  # short


def get_rxmpmp_dict(**kwargs) -> dict:
    """
    Select appropriate RXM-PMP payload definition by checking
    value of 'version' attribute (1st byte of payload).

    :param kwargs: optional payload key/value pairs
    :return: dictionary representing payload definition
    :rtype: dict
    :raises: UBXMessageError

    """

    if "version" in kwargs:
        ver = val2bytes(kwargs["version"], U1)
    elif "payload" in kwargs:
        ver = kwargs["payload"][0:1]
    else:
        raise UBXMessageError(
            "RXM-PMP message definitions must include version or payload keyword"
        )
    if ver == b"\x00":
        return UBX_PAYLOADS_SET["RXM-PMP-V0"]
    return UBX_PAYLOADS_SET["RXM-PMP-V1"]


def get_rxmrlm_dict(**kwargs) -> dict:
    """
    Select appropriate RXM-RLM payload definition by checking
    value of 'type' attribute (2nd byte of payload).

    :param kwargs: optional payload key/value pairs
    :return: dictionary representing payload definition
    :rtype: dict
    :raises: UBXMessageError

    """

    if "type" in kwargs:
        typ = val2bytes(kwargs["type"], U1)
    elif "payload" in kwargs:
        typ = kwargs["payload"][1:2]
    else:
        raise UBXMessageError(
            "RXM-RLM message definitions must include type or payload keyword"
        )
    if typ == b"\x01":
        return UBX_PAYLOADS_GET["RXM-RLM-S"]  # short
    return UBX_PAYLOADS_GET["RXM-RLM-L"]  # long


def get_cfgnmea_dict(**kwargs) -> dict:
    """
    Select appropriate payload definition version for older
    generations of CFG-NMEA message by checking payload length.

    :param kwargs: optional payload key/value pairs
    :return: dictionary representing payload definition
    :rtype: dict
    :raises: UBXMessageError

    """

    if "payload" in kwargs:
        lpd = len(kwargs["payload"])
    else:
        raise UBXMessageError(
            "CFG-NMEA message definitions must include payload keyword"
        )
    if lpd == 4:
        return UBX_PAYLOADS_GET["CFG-NMEAvX"]
    if lpd == 12:
        return UBX_PAYLOADS_GET["CFG-NMEAv0"]
    return UBX_PAYLOADS_GET["CFG-NMEA"]


def get_aopstatus_dict(**kwargs) -> dict:
    """
    Select appropriate payload definition version for older
    generations of NAV-AOPSTATUS message by checking payload length.

    :param kwargs: optional payload key/value pairs
    :return: dictionary representing payload definition
    :rtype: dict
    :raises: UBXMessageError

    """

    if "payload" in kwargs:
        lpd = len(kwargs["payload"])
    else:
        raise UBXMessageError(
            "NAV-AOPSTATUS message definitions must include payload keyword"
        )
    if lpd == 20:
        return UBX_PAYLOADS_GET["NAV-AOPSTATUS-L"]
    return UBX_PAYLOADS_GET["NAV-AOPSTATUS"]


def get_relposned_dict(**kwargs) -> dict:
    """
    Select appropriate NAV-RELPOSNED payload definition by checking
    value of 'version' attribute (1st byte of payload).

    :param kwargs: optional payload key/value pairs
    :return: dictionary representing payload definition
    :rtype: dict
    :raises: UBXMessageError

    """

    if "version" in kwargs:
        ver = val2bytes(kwargs["version"], U1)
    elif "payload" in kwargs:
        ver = kwargs["payload"][0:1]
    else:
        raise UBXMessageError(
            "NAV-RELPOSNED message definitions must include version or payload keyword"
        )
    if ver == b"\x00":
        return UBX_PAYLOADS_GET["NAV-RELPOSNED-V0"]
    return UBX_PAYLOADS_GET["NAV-RELPOSNED"]


def get_timvcocal_dict(**kwargs) -> dict:
    """
    Select appropriate TIM-VCOCAL SET payload definition by checking
    the payload length.

    :param kwargs: optional payload key/value pairs
    :return: dictionary representing payload definition
    :rtype: dict
    :raises: UBXMessageError

    """

    lpd = 1
    typ = 0
    if "type" in kwargs:
        typ = kwargs["type"]
    elif "payload" in kwargs:
        lpd = len(kwargs["payload"])
    else:
        raise UBXMessageError(
            "TIM-VCOCAL SET message definitions must include type or payload keyword"
        )
    if lpd == 1 and typ == 0:
        return UBX_PAYLOADS_SET["TIM-VCOCAL-V0"]  # stop cal
    return UBX_PAYLOADS_SET["TIM-VCOCAL"]  # cal


def get_cfgdat_dict(**kwargs) -> dict:
    """
    Select appropriate CFG-DAT SET payload definition by checking
    presence of datumNum keyword or payload length of 2 bytes.

    :param kwargs: optional payload key/value pairs
    :return: dictionary representing payload definition
    :rtype: dict

    """

    lpd = 0
    if "payload" in kwargs:
        lpd = len(kwargs["payload"])
    if lpd == 2 or "datumNum" in kwargs:
        return UBX_PAYLOADS_SET["CFG-DAT-NUM"]  # datum num set
    return UBX_PAYLOADS_SET["CFG-DAT"]  # manual datum set


def get_secsig_dict(**kwargs) -> dict:
    """
    Select appropriate SEC-SIG GET payload definition by checking
    value of 'version' attribute (1st byte of payload).

    :param kwargs: optional payload key/value pairs
    :return: dictionary representing payload definition
    :rtype: dict

    """

    if "version" in kwargs:
        ver = val2bytes(kwargs["version"], U1)
    elif "payload" in kwargs:
        ver = kwargs["payload"][0:1]
    else:
        raise UBXMessageError(
            "SEC-SIG message definitions must include version or payload keyword"
        )
    if ver == b"\x01":
        return UBX_PAYLOADS_GET["SEC-SIG-V1"]
    return UBX_PAYLOADS_GET["SEC-SIG-V2"]


def get_alpsrv_dict(**kwargs) -> dict:
    """
    Select appropriate AID-ALPSRV GET payload definition by checking
    value of 'type' attribute (2nd byte of payload).

    :param kwargs: optional payload key/value pairs
    :return: dictionary representing payload definition
    :rtype: dict

    """

    if "type" in kwargs:
        typ = val2bytes(kwargs["type"], U1)
    elif "payload" in kwargs:
        typ = kwargs["payload"][1:2]
    else:
        raise UBXMessageError(
            "AID-ALPSRV GET message definitions must include type or payload keyword"
        )
    if typ == b"\xff":
        return UBX_PAYLOADS_GET["AID-ALPSRV-SEND"]
    return UBX_PAYLOADS_GET["AID-ALPSRV-REQ"]


VARIANTS = {
    POLL: {b"\x06\x31": get_cfgtp5_dict},  # CFG-TP5
    SET: {
        b"\x13\x00": get_mga_dict,  # MGA GPS
        b"\x13\x02": get_mga_dict,  # MGA GAL
        b"\x13\x03": get_mga_dict,  # MGA BDS
        b"\x13\x05": get_mga_dict,  # MGA QZSS
        b"\x13\x06": get_mga_dict,  # MGA GLO
        b"\x13\x21": get_mga_dict,  # MGA FLASH
        b"\x13\x40": get_mga_dict,  # MGA INI
        b"\x02\x72": get_rxmpmp_dict,  # RXM-PMP
        b"\x02\x41": get_rxmpmreq_dict,  # RXM-PMREQ
        b"\x0d\x15": get_timvcocal_dict,  # TIM-VCOCAL
        b"\x06\x06": get_cfgdat_dict,  # CFG-DAT
    },
    GET: {
        b"\x0b\x32": get_alpsrv_dict,  # AID-ALPSRV
        b"\x13\x21": get_mga_dict,  # MGA FLASH
        b"\x13\x60": get_mga_dict,  # MGA ACK NAK
        b"\x02\x72": get_rxmpmp_dict,  # RXM-PMP
        b"\x02\x59": get_rxmrlm_dict,  # RXM-RLM
        b"\x06\x17": get_cfgnmea_dict,  # CFG-NMEA
        b"\x01\x60": get_aopstatus_dict,  # NAV-AOPSTATUS
        b"\x01\x3c": get_relposned_dict,  # NAV-RELPOSNED
        b"\x27\x09": get_secsig_dict,  # SEC-SIG
    },
}



"""
Special case tests for pyubx2.UBXMessage

Mainly message types which have alternative payload definitions

Created on 3 Oct 2020

*** NB: must be saved in UTF-8 format ***

@author: semuadmin
"""

# pylint: disable=line-too-long, invalid-name, missing-docstring, no-member

import unittest

from pyubx2 import UBXMessage, UBXReader, SET, GET, UBXMessageError
import pyubx2.ubxtypes_configdb as ubxcdb


class SpecialTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.cfg_rinv2 = b"\xb5b\x06\x34\x03\x00\x03\x07\x04\x4b\x8c"
        self.cfg_rinv7 = (
            b"\xb5b\x06\x34\x08\x00\x03\x01\x02\x03\x04\x05\x06\x07\x61\x40"
        )
        self.cfg_rinv0 = b"\xb5b\x06\x34\x01\x00\x03\x3e\xf4"

    def tearDown(self):
        pass

    def testCfg_Rinv2(self):
        res = UBXReader.parse(self.cfg_rinv2)
        self.assertIsInstance(res, UBXMessage)

    def testCfg_Rinv2ID(self):
        res = UBXReader.parse(self.cfg_rinv2)
        self.assertEqual(res.identity, "CFG-RINV")

    def testCfg_Rinv2Str(self):
        res = UBXReader.parse(self.cfg_rinv2)
        self.assertEqual(
            str(res), "<UBX(CFG-RINV, dump=1, binary=1, data_01=7, data_02=4)>"
        )

    def testCfg_Rinv7(self):
        res = UBXReader.parse(self.cfg_rinv7)
        self.assertIsInstance(res, UBXMessage)

    def testCfg_Rinv7ID(self):
        res = UBXReader.parse(self.cfg_rinv7)
        self.assertEqual(res.identity, "CFG-RINV")

    def testCfg_Rinv7Str(self):
        res = UBXReader.parse(self.cfg_rinv7)
        self.assertEqual(
            str(res),
            "<UBX(CFG-RINV, dump=1, binary=1, data_01=1, data_02=2, data_03=3, data_04=4, data_05=5, data_06=6, data_07=7)>",
        )

    def testCfg_Rinv0(self):
        res = UBXReader.parse(self.cfg_rinv0)
        self.assertIsInstance(res, UBXMessage)

    def testCfg_Rinv0ID(self):
        res = UBXReader.parse(self.cfg_rinv0)
        self.assertEqual(res.identity, "CFG-RINV")

    def testCfg_Rinv0Str(self):
        res = UBXReader.parse(self.cfg_rinv0)
        self.assertEqual(str(res), "<UBX(CFG-RINV, dump=1, binary=1)>")

    def testFill_RXMPMREQSET(
        self,
    ):  #  test RXM-PMREQ SET constructor long version with payload keyword
        EXPECTED_RESULT = "<UBX(RXM-PMREQ, version=0, reserved0=0, duration=67305985, backup=0, force=0, uartrx=0, extint0=0, extint1=0, spics=0)>"
        res = UBXMessage(
            "RXM",
            "RXM-PMREQ",
            SET,
            payload=b"\x00\x00\x00\x00\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04",
        )
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_RXMPMREQSET2(
        self,
    ):  #  test RXM-PMREQ SET constructor long version with version keyword
        EXPECTED_RESULT = "<UBX(RXM-PMREQ, version=0, reserved0=0, duration=67305985, backup=0, force=0, uartrx=0, extint0=0, extint1=0, spics=0)>"
        res = UBXMessage(
            "RXM",
            "RXM-PMREQ",
            SET,
            version=0,
            duration=67305985,
            flags=b"\x01\x02\x03\x04",
            wakeupSources=b"\x01\x02\x03\x04",
        )
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_RXMPMREQSET3(self):  #  test RXM-PMREQ SET constructor short version
        EXPECTED_RESULT = "<UBX(RXM-PMREQ, duration=67305985, backup=1)>"
        res = UBXMessage(
            "RXM", "RXM-PMREQ", SET, payload=b"\x01\x02\x03\x04\x02\x01\x03\x04"
        )
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_RXMPMPSETV0(
        self,
    ):  #  test RXM-PMP SET constructor with version=0 keyword
        EXPECTED_RESULT = "<UBX(RXM-PMP, version=0, reserved0=0, timeTag=0, uniqueWord1=0, uniqueWord2=0, serviceIdentifier=0, spare=0, uniqueWordBitErrors=0, userData_01=0, userData_02=0, userData_03=0, userData_04=0, userData_05=0, userData_06=0, userData_07=0, userData_08=0, userData_09=0, userData_10=0, userData_11=0, userData_12=0, userData_13=0, userData_14=0, userData_15=0, userData_16=0, userData_17=0, userData_18=0, userData_19=0, userData_20=0, userData_21=0, userData_22=0, userData_23=0, userData_24=0, userData_25=0, userData_26=0, userData_27=0, userData_28=0, userData_29=0, userData_30=0, userData_31=0, userData_32=0, userData_33=0, userData_34=0, userData_35=0, userData_36=0, userData_37=0, userData_38=0, userData_39=0, userData_40=0, userData_41=0, userData_42=0, userData_43=0, userData_44=0, userData_45=0, userData_46=0, userData_47=0, userData_48=0, userData_49=0, userData_50=0, userData_51=0, userData_52=0, userData_53=0, userData_54=0, userData_55=0, userData_56=0, userData_57=0, userData_58=0, userData_59=0, userData_60=0, userData_61=0, userData_62=0, userData_63=0, userData_64=0, userData_65=0, userData_66=0, userData_67=0, userData_68=0, userData_69=0, userData_70=0, userData_71=0, userData_72=0, userData_73=0, userData_74=0, userData_75=0, userData_76=0, userData_77=0, userData_78=0, userData_79=0, userData_80=0, userData_81=0, userData_82=0, userData_83=0, userData_84=0, userData_85=0, userData_86=0, userData_87=0, userData_88=0, userData_89=0, userData_90=0, userData_91=0, userData_92=0, userData_93=0, userData_94=0, userData_95=0, userData_96=0, userData_97=0, userData_98=0, userData_99=0, userData_100=0, userData_101=0, userData_102=0, userData_103=0, userData_104=0, userData_105=0, userData_106=0, userData_107=0, userData_108=0, userData_109=0, userData_110=0, userData_111=0, userData_112=0, userData_113=0, userData_114=0, userData_115=0, userData_116=0, userData_117=0, userData_118=0, userData_119=0, userData_120=0, userData_121=0, userData_122=0, userData_123=0, userData_124=0, userData_125=0, userData_126=0, userData_127=0, userData_128=0, userData_129=0, userData_130=0, userData_131=0, userData_132=0, userData_133=0, userData_134=0, userData_135=0, userData_136=0, userData_137=0, userData_138=0, userData_139=0, userData_140=0, userData_141=0, userData_142=0, userData_143=0, userData_144=0, userData_145=0, userData_146=0, userData_147=0, userData_148=0, userData_149=0, userData_150=0, userData_151=0, userData_152=0, userData_153=0, userData_154=0, userData_155=0, userData_156=0, userData_157=0, userData_158=0, userData_159=0, userData_160=0, userData_161=0, userData_162=0, userData_163=0, userData_164=0, userData_165=0, userData_166=0, userData_167=0, userData_168=0, userData_169=0, userData_170=0, userData_171=0, userData_172=0, userData_173=0, userData_174=0, userData_175=0, userData_176=0, userData_177=0, userData_178=0, userData_179=0, userData_180=0, userData_181=0, userData_182=0, userData_183=0, userData_184=0, userData_185=0, userData_186=0, userData_187=0, userData_188=0, userData_189=0, userData_190=0, userData_191=0, userData_192=0, userData_193=0, userData_194=0, userData_195=0, userData_196=0, userData_197=0, userData_198=0, userData_199=0, userData_200=0, userData_201=0, userData_202=0, userData_203=0, userData_204=0, userData_205=0, userData_206=0, userData_207=0, userData_208=0, userData_209=0, userData_210=0, userData_211=0, userData_212=0, userData_213=0, userData_214=0, userData_215=0, userData_216=0, userData_217=0, userData_218=0, userData_219=0, userData_220=0, userData_221=0, userData_222=0, userData_223=0, userData_224=0, userData_225=0, userData_226=0, userData_227=0, userData_228=0, userData_229=0, userData_230=0, userData_231=0, userData_232=0, userData_233=0, userData_234=0, userData_235=0, userData_236=0, userData_237=0, userData_238=0, userData_239=0, userData_240=0, userData_241=0, userData_242=0, userData_243=0, userData_244=0, userData_245=0, userData_246=0, userData_247=0, userData_248=0, userData_249=0, userData_250=0, userData_251=0, userData_252=0, userData_253=0, userData_254=0, userData_255=0, userData_256=0, userData_257=0, userData_258=0, userData_259=0, userData_260=0, userData_261=0, userData_262=0, userData_263=0, userData_264=0, userData_265=0, userData_266=0, userData_267=0, userData_268=0, userData_269=0, userData_270=0, userData_271=0, userData_272=0, userData_273=0, userData_274=0, userData_275=0, userData_276=0, userData_277=0, userData_278=0, userData_279=0, userData_280=0, userData_281=0, userData_282=0, userData_283=0, userData_284=0, userData_285=0, userData_286=0, userData_287=0, userData_288=0, userData_289=0, userData_290=0, userData_291=0, userData_292=0, userData_293=0, userData_294=0, userData_295=0, userData_296=0, userData_297=0, userData_298=0, userData_299=0, userData_300=0, userData_301=0, userData_302=0, userData_303=0, userData_304=0, userData_305=0, userData_306=0, userData_307=0, userData_308=0, userData_309=0, userData_310=0, userData_311=0, userData_312=0, userData_313=0, userData_314=0, userData_315=0, userData_316=0, userData_317=0, userData_318=0, userData_319=0, userData_320=0, userData_321=0, userData_322=0, userData_323=0, userData_324=0, userData_325=0, userData_326=0, userData_327=0, userData_328=0, userData_329=0, userData_330=0, userData_331=0, userData_332=0, userData_333=0, userData_334=0, userData_335=0, userData_336=0, userData_337=0, userData_338=0, userData_339=0, userData_340=0, userData_341=0, userData_342=0, userData_343=0, userData_344=0, userData_345=0, userData_346=0, userData_347=0, userData_348=0, userData_349=0, userData_350=0, userData_351=0, userData_352=0, userData_353=0, userData_354=0, userData_355=0, userData_356=0, userData_357=0, userData_358=0, userData_359=0, userData_360=0, userData_361=0, userData_362=0, userData_363=0, userData_364=0, userData_365=0, userData_366=0, userData_367=0, userData_368=0, userData_369=0, userData_370=0, userData_371=0, userData_372=0, userData_373=0, userData_374=0, userData_375=0, userData_376=0, userData_377=0, userData_378=0, userData_379=0, userData_380=0, userData_381=0, userData_382=0, userData_383=0, userData_384=0, userData_385=0, userData_386=0, userData_387=0, userData_388=0, userData_389=0, userData_390=0, userData_391=0, userData_392=0, userData_393=0, userData_394=0, userData_395=0, userData_396=0, userData_397=0, userData_398=0, userData_399=0, userData_400=0, userData_401=0, userData_402=0, userData_403=0, userData_404=0, userData_405=0, userData_406=0, userData_407=0, userData_408=0, userData_409=0, userData_410=0, userData_411=0, userData_412=0, userData_413=0, userData_414=0, userData_415=0, userData_416=0, userData_417=0, userData_418=0, userData_419=0, userData_420=0, userData_421=0, userData_422=0, userData_423=0, userData_424=0, userData_425=0, userData_426=0, userData_427=0, userData_428=0, userData_429=0, userData_430=0, userData_431=0, userData_432=0, userData_433=0, userData_434=0, userData_435=0, userData_436=0, userData_437=0, userData_438=0, userData_439=0, userData_440=0, userData_441=0, userData_442=0, userData_443=0, userData_444=0, userData_445=0, userData_446=0, userData_447=0, userData_448=0, userData_449=0, userData_450=0, userData_451=0, userData_452=0, userData_453=0, userData_454=0, userData_455=0, userData_456=0, userData_457=0, userData_458=0, userData_459=0, userData_460=0, userData_461=0, userData_462=0, userData_463=0, userData_464=0, userData_465=0, userData_466=0, userData_467=0, userData_468=0, userData_469=0, userData_470=0, userData_471=0, userData_472=0, userData_473=0, userData_474=0, userData_475=0, userData_476=0, userData_477=0, userData_478=0, userData_479=0, userData_480=0, userData_481=0, userData_482=0, userData_483=0, userData_484=0, userData_485=0, userData_486=0, userData_487=0, userData_488=0, userData_489=0, userData_490=0, userData_491=0, userData_492=0, userData_493=0, userData_494=0, userData_495=0, userData_496=0, userData_497=0, userData_498=0, userData_499=0, userData_500=0, userData_501=0, userData_502=0, userData_503=0, userData_504=0, fecBits=0, ebno=0, reserved1=0)>"
        res = UBXMessage("RXM", "RXM-PMP", SET, version=0)
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_RXMRLMSETV1(
        self,
    ):  #  test RXM-PMP SET constructor with version=1 keyword
        EXPECTED_RESULT = "<UBX(RXM-PMP, version=1, reserved0=0, numBytesUserData=0, timeTag=0, uniqueWord1=0, uniqueWord2=0, serviceIdentifier=0, spare=0, uniqueWordBitErrors=0, fecBits=0, ebno=0, reserved1=0)>"
        res = UBXMessage("RXM", "RXM-PMP", SET, version=1)
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_RXMRLMSETS(self):  #  test RXM-RLM GET constructor with type=1 keyword
        EXPECTED_RESULT = "<UBX(RXM-RLM, version=0, type=1, svId=0, reserved0=0, beacon=0, message=0, params=0, reserved1=0)>"
        res = UBXMessage("RXM", "RXM-RLM", GET, version=0, type=1)
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_RXMPMPSETL(self):  #  test RXM-RLM GET constructor with type=2 keyword
        EXPECTED_RESULT = "<UBX(RXM-RLM, version=0, type=2, svId=0, reserved0=0, beacon=0, message=0, params=0, reserved1=0)>"
        res = UBXMessage("RXM", "RXM-RLM", GET, version=0, type=2)
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_ESFMEASSET(
        self,
    ):  #  test ESF_MEAS GET constructor with calibTtagValid = 0
        EXPECTED_RESULT = "<UBX(ESF-MEAS, timeTag=0, timeMarkSent=0, timeMarkEdge=0, calibTtagValid=0, numMeas=0, id=0)>"
        res = UBXMessage(
            "ESF", "ESF-MEAS", GET, timeTag=0, flags=b"\x00\x00", parsebitfield=0
        )
        res2 = UBXReader.parse(res.serialize())
        self.assertEqual(str(res2), EXPECTED_RESULT)

    def testFill_ESFMEASSETCT(
        self,
    ):  #  test ESF_MEAS GET constructor with calibTtagValid = 1
        EXPECTED_RESULT = "<UBX(ESF-MEAS, timeTag=0, timeMarkSent=0, timeMarkEdge=0, calibTtagValid=1, numMeas=0, id=0)>"
        res = UBXMessage(
            "ESF", "ESF-MEAS", GET, timeTag=0, flags=b"\x18\x00", parsebitfield=0
        )
        res2 = UBXReader.parse(res.serialize())
        self.assertEqual(str(res2), EXPECTED_RESULT)

    def testFill_MGABDSSET(
        self,
    ):  #  test MGA-BDS-UTC SET constructor using attribute keywords, very large scaling factors
        EXPECTED_RESULT = "<UBX(MGA-BDS-UTC, type=5, version=0, reserved0=0, a0UTC=1.2218952178955079e-09, a1UTC=1.2949641359227826e-15, dtLS=0, reserved1=0, wnRec=23, wnLSF=41, dN=0, dtLSF=0, reserved2=0)>"
        res = UBXMessage(
            b"\x13",
            b"\x03",
            SET,
            type=5,
            a0UTC=1.312 * 2**-30,
            a1UTC=1.458 * 2**-50,
            wnRec=23,
            wnLSF=41,
        )
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_RELPOSNED_v0(self):  # test NAV-RELPOSNED V0 constructor
        EXPECTED_RESULT = "<UBX(NAV-RELPOSNED, version=0, reserved1=0, refStationID=0, iTOW=23:59:42, relPosN=1, relPosE=2, relPosD=3, reserved2=0, accN=0, accE=0, accD=0, gnssFixOK=0, diffSoln=0, relPosValid=0, carrSoln=0, isMoving=0, refPosMiss=0, refObsMiss=0, relPosHeadingValid=0, relPosNormalized=0)>"
        EXPECTED_RESULT2 = "<UBX(NAV-RELPOSNED, version=0, reserved1=0, refStationID=0, iTOW=23:59:42, relPosN=1, relPosE=2, relPosD=3, reserved2=0, accN=0.0, accE=0.0, accD=0.0, gnssFixOK=0, diffSoln=0, relPosValid=0, carrSoln=0, isMoving=0, refPosMiss=0, refObsMiss=0, relPosHeadingValid=0, relPosNormalized=0)>"
        res = UBXMessage(
            "NAV",
            "NAV-RELPOSNED",
            GET,
            version=0,
            relPosN=1,
            relPosE=2,
            relPosD=3,
        )
        res2 = UBXMessage("NAV", "NAV-RELPOSNED", GET, payload=res.payload)
        self.assertEqual(str(res), EXPECTED_RESULT)
        # self.assertEqual(str(res2), EXPECTED_RESULT)

    def testFill_RELPOSNED_v1(self):  # test NAV-RELPOSNED V1 constructor
        EXPECTED_RESULT = "<UBX(NAV-RELPOSNED, version=1, reserved0=0, refStationID=0, iTOW=23:59:42, relPosN=1, relPosE=2, relPosD=3, relPosLength=0, relPosHeading=0, reserved1=0, accN=0, accE=0, accD=0, accLength=0, accHeading=0, reserved2=0, gnssFixOK=0, diffSoln=0, relPosValid=0, carrSoln=0, isMoving=0, refPosMiss=0, refObsMiss=0, relPosHeadingValid=0, relPosNormalized=0)>"
        EXPECTED_RESULT2 = "<UBX(NAV-RELPOSNED, version=1, reserved0=0, refStationID=0, iTOW=23:59:42, relPosN=1, relPosE=2, relPosD=3, relPosLength=0, relPosHeading=0.0, reserved1=0, accN=0.0, accE=0.0, accD=0.0, accLength=0.0, accHeading=0.0, reserved2=0, gnssFixOK=0, diffSoln=0, relPosValid=0, carrSoln=0, isMoving=0, refPosMiss=0, refObsMiss=0, relPosHeadingValid=0, relPosNormalized=0)>"
        res = UBXMessage(
            "NAV",
            "NAV-RELPOSNED",
            GET,
            version=1,
            relPosN=1,
            relPosE=2,
            relPosD=3,
        )
        res2 = UBXMessage("NAV", "NAV-RELPOSNED", GET, payload=res.payload)
        self.assertEqual(str(res), EXPECTED_RESULT)
        # self.assertEqual(str(res2), EXPECTED_RESULT)

    def testFill_AOPSTATUSL(self):  # testNAV-AOPSTATUS M6 constructor
        EXPECTED_RESULT = "<UBX(NAV-AOPSTATUS, iTOW=18:41:27.985000, config=1, status=2, reserved0=0, reserved1=0, avail=67305985, reserved2=67305985, reserved3=67305985)>"
        res = UBXMessage(
            "NAV",
            "NAV-AOPSTATUS",
            GET,
            payload=b"\x01\x02\x03\x04\x01\x02\x00\x00\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04",
        )
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_AOPSTATUS(self):  # testNAV-AOPSTATUS M8 constructor
        EXPECTED_RESULT = "<UBX(NAV-AOPSTATUS, iTOW=18:41:27.985000, aopCfg=1, status=2, reserved1=42649378395939397566720)>"
        res = UBXMessage(
            "NAV",
            "NAV-AOPSTATUS",
            GET,
            payload=b"\x01\x02\x03\x04\x01\x02\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09",
        )
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_TIMVCOCAL_v0(self):  # test TIM-VCOCAL-V0 constructor
        EXPECTED_RESULT = "<UBX(TIM-VCOCAL, type=0)>"
        res = UBXMessage(
            "TIM",
            "TIM-VCOCAL",
            SET,
            type=0,
        )
        res2 = UBXMessage("TIM", "TIM-VCOCAL", SET, payload=res.payload)
        self.assertEqual(str(res), EXPECTED_RESULT)
        self.assertEqual(str(res2), EXPECTED_RESULT)

    def testFill_TIMVCOCAL(self):  # test TIM-VCOCAL constructor
        EXPECTED_RESULT = "<UBX(TIM-VCOCAL, type=2, version=0, oscId=0, srcId=0, reserved1=0, raw0=0, raw1=0, maxStepSize=2)>"
        res = UBXMessage(
            "TIM",
            "TIM-VCOCAL",
            SET,
            type=2,
            maxStepSize=2,
        )
        res2 = UBXMessage("TIM", "TIM-VCOCAL", SET, payload=res.payload)
        self.assertEqual(str(res), EXPECTED_RESULT)
        self.assertEqual(str(res2), EXPECTED_RESULT)

    def testFill_CFGDAT(self):  # test CFG-DAT constructor
        EXPECTED_RESULT = "<UBX(CFG-DAT, majA=15.2, flat=23.4, dX=4.6, dY=7.2, dZ=15.7, rotX=123.3, rotY=18.4, rotZ=43.5, scale=2.3)>"
        res = UBXMessage(
            "CFG",
            "CFG-DAT",
            SET,
            majA=15.2,
            flat=23.4,
            dX=4.6,
            dY=7.2,
            dZ=15.7,
            rotX=123.3,
            rotY=18.4,
            rotZ=43.5,
            scale=2.3,
        )
        self.assertEqual(str(res), EXPECTED_RESULT)
        res2 = UBXMessage("CFG", "CFG-DAT", SET, payload=res.payload)
        self.assertAlmostEqual(res2.majA, 15.2, 5)
        self.assertAlmostEqual(res2.dX, 4.6, 5)
        self.assertAlmostEqual(res2.scale, 2.3, 5)

    def testFill_CFGDAT_NUM(self):  # test CFG-DAT-NUM constructor
        EXPECTED_RESULT = "<UBX(CFG-DAT, datumNum=7)>"
        res = UBXMessage(
            "CFG",
            "CFG-DAT",
            SET,
            datumNum=7,
        )
        res2 = UBXMessage("CFG", "CFG-DAT", SET, payload=res.payload)
        self.assertEqual(str(res), EXPECTED_RESULT)
        self.assertEqual(str(res2), EXPECTED_RESULT)

    def testConfigSet(self):  # test creation of CFG-VALSET message with single key
        cfgData = [("CFG_UART1_BAUDRATE", 9600)]
        res = UBXMessage.config_set(ubxcdb.SET_LAYER_RAM, ubxcdb.TXN_NONE, cfgData)
        self.assertEqual(
            str(res),
            "<UBX(CFG-VALSET, version=0, ram=1, bbr=0, flash=0, action=0, reserved0=0, CFG_UART1_BAUDRATE=9600)>",
        )

    def testConfigSet2(
        self,
    ):  # test creation of CFG-VALSET message with multiple keys as transaction
        cfgData = [("CFG_UART1_BAUDRATE", 9600), (0x40530001, 115200)]
        res = UBXMessage.config_set(ubxcdb.SET_LAYER_BBR, ubxcdb.TXN_START, cfgData)
        self.assertEqual(
            str(res),
            "<UBX(CFG-VALSET, version=1, ram=0, bbr=1, flash=0, action=1, reserved0=0, CFG_UART1_BAUDRATE=9600, CFG_UART2_BAUDRATE=115200)>",
        )

    def testConfigDel(self):  # test creation of CFG-VALSET message with single key
        keys = [
            "CFG_UART1_BAUDRATE",
        ]
        res = UBXMessage.config_del(ubxcdb.SET_LAYER_BBR, ubxcdb.TXN_NONE, keys)
        self.assertEqual(
            str(res),
            "<UBX(CFG-VALDEL, version=0, bbr=1, flash=0, action=0, reserved0=0, keys_01=1079115777)>",
        )

    def testConfigDel2(
        self,
    ):  # test creation of CFG-VALSET message with multiples keys as transaction
        keys = ["CFG_UART1_BAUDRATE", 0x40530001]
        res = UBXMessage.config_del(ubxcdb.SET_LAYER_FLASH, ubxcdb.TXN_START, keys)
        self.assertEqual(
            str(res),
            "<UBX(CFG-VALDEL, version=1, bbr=0, flash=1, action=1, reserved0=0, keys_01=1079115777, keys_02=1079181313)>",
        )

    def testConfigPoll(self):  # test creation of CFG-VALGET message with multiple keys
        keys = ["CFG_UART1_BAUDRATE", 0x40530001]
        res = UBXMessage.config_poll(ubxcdb.POLL_LAYER_FLASH, 0, keys)
        self.assertEqual(
            str(res),
            "<UBX(CFG-VALGET, version=0, layer=2, position=0, keys_01=1079115777, keys_02=1079181313)>",
        )

    def testESFMEASSET0(self):  # test generation fo ESF-MEAS with calibTtagValid=0
        EXPECTED_RESULT = "<UBX(ESF-MEAS, timeTag=12345000, timeMarkSent=0, timeMarkEdge=0, calibTtagValid=0, numMeas=1, id=0, dataField_01=188, dataType_01=11)>"
        msg = UBXMessage(
            "ESF",
            "ESF-MEAS",
            SET,
            timeTag=12345000,
            timeMarkSent=0,
            timeMarkEdge=0,
            calibTtagValid=0,
            numMeas=1,
            dataField_01=188,
            dataType_01=11,
        )
        self.assertEqual(str(msg), EXPECTED_RESULT)

    def testESFMEASSET1(self):  # test generation fo ESF-MEAS with calibTtagValid=1
        EXPECTED_RESULT = "<UBX(ESF-MEAS, timeTag=12345000, timeMarkSent=0, timeMarkEdge=0, calibTtagValid=1, numMeas=2, id=0, dataField_01=223, dataType_01=9, dataField_02=118, dataType_02=11, dataField_03=12345000, dataType_03=0)>"
        msg = UBXMessage(
            "ESF",
            "ESF-MEAS",
            SET,
            timeTag=12345000,
            timeMarkSent=0,
            timeMarkEdge=0,
            calibTtagValid=1,
            numMeas=2,
            dataField_01=223,
            dataType_01=9,
            dataField_02=118,
            dataType_02=11,
            dataField_03=12345000,
            dataType_03=0,
        )
        self.assertEqual(str(msg), EXPECTED_RESULT)

    def testACKACKNK(self):  # test ACK-ACK of unknown message class
        EXPECTED_RESULT = "<UBX(ACK-ACK, clsID=b'w', msgID=b'w\\x88')>"
        msg = UBXMessage("ACK", "ACK-ACK", GET, clsID=0x77, msgID=0x88)
        self.assertEqual(str(msg), EXPECTED_RESULT)

    def testCFGMSGNK(self):  # test CFG-MSG of unknown message class
        EXPECTED_RESULT = "<UBX(CFG-MSG, msgClass=b'w', msgID=b'w\\x88', rateDDC=0, rateUART1=4, rateUART2=0, rateUSB=0, rateSPI=0, reserved=0)>"
        msg = UBXMessage("CFG", "CFG-MSG", GET, msgClass=0x77, msgID=0x88, rateUART1=4)
        self.assertEqual(str(msg), EXPECTED_RESULT)

    def testUnknownClassID(self):  # test for unknown class and id
        EXPECTED_RESULT = "<UBX(UNKNOWN-5566-NOMINAL, payload=b'\\x33\\x44')>"
        msg = UBXReader.parse(b"\xb5b\x55\x66\x02\x00\x33\x44\x34\xae")
        # print(msg)
        self.assertEqual(str(msg), EXPECTED_RESULT)

    def testUnknownID(self):  # test for known class, unknown id
        EXPECTED_RESULT = "<UBX(DBG-0c66-NOMINAL, payload=b'\\x55\\x66')>"
        msg = UBXReader.parse(b"\xb5b\x0c\x66\x02\x00\x55\x66\x2f\x5e")
        # print(msg)
        self.assertEqual(str(msg), EXPECTED_RESULT)

    def testAIDALPSRVREQ(self):
        EXPECTED_RESULT = "<UBX(AID-ALPSRV, idSize=60, type=253, ofs=23, size=24, fileId=12, dataSize=6, id1=1, id2=2, id3=3)>"
        msg = UBXMessage(
            "AID",
            "AID-ALPSRV",
            GET,
            idSize=0x3C,
            type=0xFD,
            ofs=23,
            size=24,
            fileId=12,
            dataSize=6,
            id1=1,
            id2=2,
            id3=3,
        )
        # print(msg)
        self.assertEqual(str(msg), EXPECTED_RESULT)

    def testAIDALPSRVSEND(self):
        EXPECTED_RESULT = "<UBX(AID-ALPSRV, idSize=60, type=255, ofs=23, size=6, fileId=12, data_01=1, data_02=2, data_03=3, data_04=0, data_05=0, data_06=0)>"
        msg = UBXMessage(
            "AID",
            "AID-ALPSRV",
            GET,
            idSize=0x3C,
            type=0xFF,
            ofs=23,
            size=6,
            fileId=12,
            data_01=1,
            data_02=2,
            data_03=3,
        )
        # print(msg)
        self.assertEqual(str(msg), EXPECTED_RESULT)

    def testAIDALPSRVNOTYPE(self):
        with self.assertRaisesRegex(
            UBXMessageError,
            "AID-ALPSRV GET message definitions must include type or payload keyword",
        ):
            msg = UBXMessage(
                "AID",
                "AID-ALPSRV",
                GET,
                idSize=0x3C,
                ofs=23,
                size=6,
                fileId=12,
                data_01=1,
                data_02=2,
                data_03=3,
            )


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
UBX Custom Exception Types.

Created on 27 Sep 2020

:author: semuadmin
:copyright: SEMU Consulting © 2020
:license: BSD 3-Clause
"""


class ParameterError(Exception):
    """Parameter Error Class."""


class GNSSStreamError(Exception):
    """Generic Stream Error Class."""


class UBXParseError(Exception):
    """
    UBX Parsing error.
    """


class UBXStreamError(Exception):
    """
    UBX Streaming error.
    """


class UBXMessageError(Exception):
    """
    UBX Undefined message class/id.
    Essentially a prompt to add missing payload types to UBX_PAYLOADS.
    """


class UBXTypeError(Exception):
    """
    UBX Undefined payload attribute type.
    Essentially a prompt to fix incorrect payload definitions to UBX_PAYLOADS.
    """



"""
Bitfield parse method tests for pyubx2.UBXMessage

Created on 16 Oct 2021

*** NB: must be saved in UTF-8 format ***

@author: semuadmin
"""
# pylint: disable=line-too-long, invalid-name, missing-docstring, no-member

import unittest

from pyubx2 import UBXMessage, UBXReader, GET


class ParseTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.nav_sat2 = b"\xb5b\x015\x14\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00+*\xf0\x00\xfd\xff\x07\x07\x00\x00\x9bt"  # qualityInd = 7, orbitSource = 7
        self.nav_sat3 = b"\xb5b\x015\x14\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00+*\xf0\x00\xfd\xff\x03\x04\x02\x00\x96_"  # rtcmCorrUsed = 1
        self.nav_sat4 = b"\xb5b\x015\x14\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00+*\xf0\x00\xfd\xff\x0a\x15\x5f\x00\x0bh"  # svUsed = 1, almUsed = 1

    def tearDown(self):
        pass

    def testNavSat2(
        self,
    ):  # check X4 bitfield correctly parsed (remember little-endian)
        res = UBXReader.parse(self.nav_sat2)
        self.assertEqual(
            str(res),
            "<UBX(NAV-SAT, iTOW=23:59:42, version=1, numSvs=1, reserved0=0, gnssId_01=GPS, svId_01=0, cno_01=43, elev_01=42, azim_01=240, prRes_01=-0.3, qualityInd_01=7, svUsed_01=0, health_01=0, diffCorr_01=0, smoothed_01=0, orbitSource_01=7, ephAvail_01=0, almAvail_01=0, anoAvail_01=0, aopAvail_01=0, sbasCorrUsed_01=0, rtcmCorrUsed_01=0, slasCorrUsed_01=0, spartnCorrUsed_01=0, prCorrUsed_01=0, crCorrUsed_01=0, doCorrUsed_01=0)>",
        )

    def testNavSat3(
        self,
    ):  # check X4 bitfield correctly parsed (remember little-endian)
        res = UBXReader.parse(self.nav_sat3)
        self.assertEqual(
            str(res),
            "<UBX(NAV-SAT, iTOW=23:59:42, version=1, numSvs=1, reserved0=0, gnssId_01=GPS, svId_01=0, cno_01=43, elev_01=42, azim_01=240, prRes_01=-0.3, qualityInd_01=3, svUsed_01=0, health_01=0, diffCorr_01=0, smoothed_01=0, orbitSource_01=4, ephAvail_01=0, almAvail_01=0, anoAvail_01=0, aopAvail_01=0, sbasCorrUsed_01=0, rtcmCorrUsed_01=1, slasCorrUsed_01=0, spartnCorrUsed_01=0, prCorrUsed_01=0, crCorrUsed_01=0, doCorrUsed_01=0)>",
        )

    def testNavSat4(
        self,
    ):  # check X4 bitfield correctly parsed (remember little-endian)
        res = UBXReader.parse(self.nav_sat4)
        self.assertEqual(
            str(res),
            "<UBX(NAV-SAT, iTOW=23:59:42, version=1, numSvs=1, reserved0=0, gnssId_01=GPS, svId_01=0, cno_01=43, elev_01=42, azim_01=240, prRes_01=-0.3, qualityInd_01=2, svUsed_01=1, health_01=0, diffCorr_01=0, smoothed_01=0, orbitSource_01=5, ephAvail_01=0, almAvail_01=1, anoAvail_01=0, aopAvail_01=0, sbasCorrUsed_01=1, rtcmCorrUsed_01=1, slasCorrUsed_01=1, spartnCorrUsed_01=1, prCorrUsed_01=1, crCorrUsed_01=0, doCorrUsed_01=1)>",
        )

    def testNavSat5(self):  # check message bytes match original byte stream
        res = UBXReader.parse(self.nav_sat4)
        self.assertEqual(res.serialize(), self.nav_sat4)

    def testNavSat6(
        self,
    ):  # check message correctly constructed from individual bit flags
        res = UBXMessage(
            "NAV",
            "NAV-SAT",
            GET,
            version=1,
            numSvs=1,
            gnssId_01=0,
            svId_01=0,
            cno_01=43,
            elev_01=42,
            azim_01=240,
            prRes_01=-0.30000000000000004,
            qualityInd_01=2,
            svUsed_01=1,
            health_01=0,
            diffCorr_01=0,
            smoothed_01=0,
            orbitSource_01=5,
            ephAvail_01=0,
            almAvail_01=1,
            anoAvail_01=0,
            aopAvail_01=0,
            sbasCorrUsed_01=1,
            rtcmCorrUsed_01=1,
            slasCorrUsed_01=1,
            spartnCorrUsed_01=1,
            prCorrUsed_01=1,
            crCorrUsed_01=0,
            doCorrUsed_01=1,
        )
        self.assertEqual(
            str(res),
            "<UBX(NAV-SAT, iTOW=23:59:42, version=1, numSvs=1, reserved0=0, gnssId_01=GPS, svId_01=0, cno_01=43, elev_01=42, azim_01=240, prRes_01=-0.30000000000000004, qualityInd_01=2, svUsed_01=1, health_01=0, diffCorr_01=0, smoothed_01=0, orbitSource_01=5, ephAvail_01=0, almAvail_01=1, anoAvail_01=0, aopAvail_01=0, sbasCorrUsed_01=1, rtcmCorrUsed_01=1, slasCorrUsed_01=1, spartnCorrUsed_01=1, prCorrUsed_01=1, crCorrUsed_01=0, doCorrUsed_01=1)>",
        )
        self.assertEqual(res.serialize(), self.nav_sat4)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
UBX Protocol attribute value decode constants.

Created on 26 Aug 2023

Information sourced from public domain u-blox Interface Specifications © 2013-2021, u-blox AG

:author: semuadmin
"""

GNSSLIST = {
    0: "GPS",
    1: "SBAS",
    2: "Galileo",
    3: "BeiDou",
    4: "IMES",
    5: "QZSS",
    6: "GLONASS",
    7: "NAVIC",
}
"""GNSS code"""

# UBX-CFG-DGNSS
DGNSMODE = {
    2: "RTK float",
    3: "RTK fixed",
}
"""Differential correction mode from UBX-CFG-DGNSS"""

# UBX-CFG-GEOFENCE
CONFLVL = {
    0: "no confidence required",
    1: "68%",
    2: "95%",
    3: "99.7%",
    4: "99.99%",
}
"""Confidence level from UBX-CFG-GEOFENCE"""

# UBX-CFG-GNSS
# key is (gnssId, sigCfMask)
SIGCFMASK = {
    (0, 0x01): "GPS L1C/A",
    (0, 0x10): "GPS L2C",
    (0, 0x20): "GPS L5",
    (1, 0x01): "SBAS L1C/A",
    (2, 0x01): "Galileo E1",
    (2, 0x10): "Galileo E5a",
    (2, 0x20): "Galileo E5b",
    (3, 0x01): "BeiDou B1I",
    (3, 0x10): "BeiDou B2I",
    (3, 0x80): "BeiDou B2A",
    (4, 0x01): "IMES L1",
    (5, 0x01): "QZSS L1C/A",
    (5, 0x04): "QZSS L1S",
    (5, 0x10): "QZSS L2C",
    (5, 0x20): "QZSS L5",
    (6, 0x01): "GLONASS L1",
    (6, 0x10): "GLONASS L2",
}
"""
Signal & carrier frequency code from UBX-CFG_GNSS
 - key is (gnssId, sigCfMask)
"""

# UBX-CFG-INF
PROTOCOLID = {
    0: "UBX",
    1: "NMEA",
}
"""Protocol from UBX-CFG-INF"""

# UBX-CFG-NAV5
DYNMODEL = {
    0: "portable",
    2: "stationary",
    3: "pedestrian",
    4: "automotive",
    5: "sea",
    6: "airborne with <1g acceleration",
    7: "airborne with <2g acceleration",
    8: "airborne with <4g acceleration",
    9: "wrist-worn watch",  # (not supported for protocol versions less than 18.00)
    10: "motorbike",  # (supported for protocol versions 19.20, and 35.10)
    11: "robotic lawn mower",  # (supported for protocol versions 33.21)
    12: "electric kick scooter",  # (supported for protocol versions 33.21, and 35.10)
}
"""Dynamic model from UBX-CFG-NAV5"""

# UBX-CFG-NAV5
FIXMODE = {
    1: "2D only",
    2: "3D only",
    3: "auto 2D/3D",
}
"""Fix mode from UBX-CFG-NAV5"""

# UBX-CFG-NMEA
NMEAVERSION = {
    0x4B: "NMEA version 4.11",  # (not available in all products)
    0x41: "NMEA version 4.10",  # (not available in all products)
    0x40: "NMEA version 4.0",  # (not available in all products)
    0x23: "NMEA version 2.3",
    0x21: "NMEA version 2.1",
}
"""NMEA version from UBX-CFG-NMEA"""

# UBX-CFG-NMEA
SVNUMBERING = {
    0: "Strict - Satellites are not output",
    1: "Extended - Use proprietary numbering",  # (see Satellite Numbering)
}
"""Space vehicle numbering from UBX-CFG-NMEA"""

# UBX-CFG-NMEA
MAINTALKERID = {
    0: "Main Talker ID is not overridden",
    1: "Set main Talker ID to 'GP'",
    2: "Set main Talker ID to 'GL'",
    3: "Set main Talker ID to 'GN'",
    4: "Set main Talker ID to 'GA'",  # (not supported for protocol versions less than 15.00)
    5: "Set main Talker ID to 'GB'",  # (not supported for protocol versions less than 15.00)
    6: "Set main Talker ID to 'GQ'",  # (available in NMEA 4.11 and later)
}
"""Main talker ID from UBX-CFG-NMEA"""

# UBX-CFG-NMEA
GSVTALKERID = {
    0: "Use GNSS-specific Talker ID (as defined byNMEA)",
    1: "Use the main Talker ID",
}
"""GSV talker ID from UBX-CFG-NMEA"""

# UBX-CFG-ODO
ODOPROFILE = {0: "running", 1: "cycling", 2: "swimming", 3: "car", 4: "custom"}
"""Odometer profile from UBX-CFG-ODO"""

# UBX-CFG-PRT
CHARLEN = {
    0x00: "5bit",  # (not supported)
    0x01: "6bit",  # (not supported)
    0x10: "7bit",  # (supported only with parity)
    0x11: "8bit",
}
"""Serial port character length from UBX-CFG-PRT"""

PARITY = {
    "000": "Even parity",
    "001": "Odd parity",
    "10X": "No parity",
    "X1X": "Reserved",
}
"""Serial port parity from UBX-CFG-PRT"""

NSTOPBITS = {
    0x00: "1 Stop bit",
    0x01: "1.5 Stop bit",
    0x10: "2 Stop bit",
    0x11: "0.5 Stop bit",
}
"""Serial port stop bits from IBX-CFG-PRT"""

POL = {
    0: "High-active",
    1: "Low-active",
}
"""Serial port polarity from UBX-CFG-PRT"""

SPIMODE = {
    0x00: "SPI Mode 0: CPOL = 0, CPHA = 0",
    0x01: "SPI Mode 1: CPOL = 0, CPHA = 1",
    0x10: "SPI Mode 2: CPOL = 1, CPHA = 0",
    0x11: "SPI Mode 3: CPOL = 1, CPHA = 1",
}
"""SPI mode from UBX-CFG-PRT"""

# UBX-CFG-PWR
STATE = {
    0x52554E20: "GNSS running",
    0x53544F50: "GNSS stopped",
    0x42434B50: "Software backup",  # USB interface will be disabled, other wakeup source is needed.
}
"""Power state from UBX-CFG-PWR"""

# UBX-CFG-RATE
TIMEREF = {
    0: "UTC time",
    1: "GPS time",
    2: "GLONASS time",  # (not supported for protocol versions less than 18.00)
    3: "BeiDou time",  # (not supported for protocol versions less than 18.00)
    4: "Galileo time",  # (not supported for protocol versions less than 18.00)
    5: "NavIC time",  # (not supported for protocol versions less than 29.00)
}
"""Time reference from UBX-CFG-RATE"""

# UBX-CFG-RST
NAVBBRMASK = {
    0x0000: "Hot start",
    0x0001: "Warm start",
    0xFFFF: "Cold start",
}
"""Navigation battery-backed RAM mask from UBX-CFG-RST"""

RESETMODE = {
    0x00: "Hardware reset (watchdog) immediately",
    0x01: "Controlled software reset",
    0x02: "Controlled software reset (GNSS only)",
    0x04: "Hardware reset (watchdog) after shutdown",
    0x08: "Controlled GNSS stop",
    0x09: "Controlled GNSS start",
}
"""Reset mode from UBX-CFG-RST"""

# UBX-CFG-TMODE3
MODE = {
    0: "Disabled",
    1: "Survey In",
    2: "Fixed Mode",  # (true ARP position information required)
}
"""Timing mode from UBX-CFG-TMODE3"""

# UBX-CFG-TP5
GRIDUTCGNSS = {
    0: "UTC",
    1: "GPS",
    2: "GLONASS",
    3: "BeiDou",
    4: "Galileo",  # (not supported for protocol versions less than 18.00)
}
"""Timing Grid UTC GNSS source from UBX-CFG-TP5"""

# UBX-MON-COMMS
PROTIDS = {
    0: "UBX",
    1: "NMEA",
    2: "RTCM2",
    5: "RTCM3",
    6: "SPARTN",
    0xFF: "No protocol reported",
}
"""Protocol IDs from UBX-MON-COMMS"""

# UBX-MON-HW, UBX-MON-RF
ASTATUS = {
    0: "INIT",
    1: "DONTKNOW",
    2: "OK",
    3: "SHORT",
    4: "OPEN",
}
"""Antenna status from UBX-MON-HW, UBX-MON-RF"""

APOWER = {
    0: "OFF",
    1: "ON",
    2: "DONTKNOW",
}
"""Antenna power from UBX-MON-HW, UBX-MON-RF"""

JAMMINGSTATE = {
    0: "unknown or feature disabled",
    1: "ok - no significant jamming",
    2: "warning - interference visible but fix OK",
    3: "critical - interference visible and no fix",
}
"""Jamming state from UBX-MON-HW, UBX-MON-RF"""

# UBX-MON-SYS
BOOTTYPE = {
    0: "Unknown",
    1: "Cold Start",
    2: "Watchdog",
    3: "Hardware reset",
    4: "Hardware backup",
    5: "Software backup",
    6: "Software reset",
    7: "VIO fail",
    8: "VDD_X fail",
    9: "VDD_RF fail",
    10: "V_CORE_HIGH fail",
}
"""Boot type from UBX-MON-SYS"""

# UBX-NAV-GEOFENCE
GEOFENCE_STATUS = {
    0: "Geofencing not available or not reliable",
    1: "Geofencing active",
}
"""Geofence status from UBX-NAV-GEOFENCE"""

COMBSTATE = {
    0: "unknown",
    1: "inside",
    2: "outside",
}
"""Comb state from UBX-NAV-GEOFENCE"""

# UBX-NAV-ORB
HEALTH = {
    0: "unknown",
    1: "healthy",
    2: "not healthy",
}
"""Space vehicle health from UBX-NAV-ORB"""

VISIBILITY = {
    0: "unknown",
    1: "below horizon",
    2: "above horizon",
    3: "above elevation mask",
}
"""Space vehicle visibility from UBX-NAV-ORB"""

# UBX-NAV-PL
PLPOSFRAME = PLVELFRAME = {
    0: "Invalid (not possible to calculate frameconversion)",
    1: "North-East-Down",
    2: "Longitudinal-Lateral-Vertical",
    3: "HorizSemiMajorAxis-HorizSemiMinorAxis-Vertical",
}
"""Position protection level from UBX-NAV-PL"""

# UBX-NAV-PVT
GPSFIX = FIXTYPE = {
    0: "NO FIX",
    1: "DR",
    2: "2D",
    3: "3D",
    4: "GPS + DR",
    5: "TIME ONLY",
}
"""Fix type from UBX-NAV-PVT"""

PSMSTATE = {
    0: "PSM is not active",
    1: "Enabled",
    2: "Acquisition",
    3: "Tracking",
    4: "Power Optimized Tracking",
    5: "Inactive",
}
"""Power save mode state from UBX-NAV-PVT"""

CARRSOLN = {
    0: "NO RTK",
    1: "RTK FLOAT",
    2: "RTK FIXED",
}
"""Carrier phase range solution from UBX-NAV-PVT"""

LASTCORRECTIONAGE = {
    0: 0,
    1: 1,
    2: 2,
    3: 5,
    4: 10,
    5: 15,
    6: 20,
    7: 30,
    8: 45,
    9: 60,
    10: 90,
    11: 120,
}
"""Last DGPS correction age from UBX-NAV-PVT"""

# UBX-NAV-SAT
QUALITYIND = {
    0: "no signal",
    1: "searching signal",
    2: "signal acquired",
    3: "signal detected but unusable",
    4: "code locked and time synchronized",
    5: "code and carrier locked and time synchronized",
    6: "code and carrier locked and time synchronized",
    7: "code and carrier locked and time synchronized",
}
"""Signal quality indicator from UBX-NAV-SAT"""

ORBITSOURCE = {
    0: "no orbit information is available for this SV",
    1: "ephemeris is used",
    2: "almanac is used",
    3: "AssistNow Offline orbit is used",
    4: "AssistNow Autonomous orbit is used",
    5: "other orbit information is used",
    6: "other orbit information is used",
    7: "other orbit information is used",
}
"""Orbit source from UBX-NAV-SAT"""

# UBX-NAV-SBAS
SBASMODE = {
    0: "Disabled",
    1: "Enabled integrity",
    3: "Enabled test mode",
}
"""SBAS (satellite-based augmentation system) mode from UBX-NAV-SBAS"""

SBASSYS = {
    -1: "Unknown",
    0: "WAAS",
    1: "EGNOS",
    2: "MSAS",
    3: "GAGAN",
    16: "GPS",
}
"""SBAS system from UBX-NAV-SBAS"""

SBASINTEGRITYUSED = {
    0: "Unknown",
    1: "Integrity information is not available or SBAS integrity is not enabled",
    2: "Receiver uses only GPS satellites for which integrity information is available",
}
"""SBAS integrity used indicator from UBX-NAV-SBAS"""

# UBX-NAV-SIG
CORRSOURCE = {
    0: "NONE",
    1: "SBAS",
    2: "BeiDou",
    3: "RTCM2",
    4: "RTCM3 OSR",
    5: "RTCM3 SSR",
    6: "QZSS SLAS",
    7: "SPARTN",
    8: "CLAS",
}
"""DGPS correction source from UBX-NAV-SIG"""

IONOMODEL = {
    0: "none",
    1: "Klobuchar GPS",
    2: "SBAS",
    3: "Klobuchar BeiDou",
    4: "Dual Frequency",
}
"""Ionospheric model type from UBX-NAV-SIG"""

# key is (gnssId, sigId)
SIGID = {
    (0, 0): "GPS L1C/A",
    (0, 3): "GPS L2 CL",
    (0, 4): "GPS L2 CM",
    (0, 6): "GPS L5 I",
    (0, 7): "GPS L5 Q",
    (1, 0): "SBAS L1C/A",
    (2, 0): "Galileo E1 C",
    (2, 1): "Galileo E1 B",
    (2, 3): "Galileo E5 al",
    (2, 4): "Galileo E5 aQ",
    (2, 5): "Galileo E5 bI",
    (2, 6): "Galileo E5 bQ",
    (3, 0): "BeiDou B1I D1",
    (3, 1): "BeiDou B1I D2",
    (3, 2): "BeiDou B2I D1",
    (3, 3): "BeiDou B2I D2",
    (3, 5): "BeiDou B1C",
    (3, 7): "BeiDou B2a",
    (5, 0): "QZSS L1C/A",
    (5, 1): "QZSS L1S",
    (5, 4): "QZSS L2 CM",
    (5, 5): "QZSS L2 CL",
    (5, 8): "QZSS L5 I",
    (5, 9): "QZSS L5 Q",
    (6, 0): "GLONASS L1 OF",
    (6, 2): "GLONASS L2 OF",
    (7, 0): "NavIC L5 A",
}
"""
GNSS, Signal ID from UBX-NAV-SIG
 - key is (gnssId, sigId)
"""

# UBX-NAV-STATUS
SPOOFDETSTATE = {
    0: "unknown or deactivated",
    1: "no spoofing indicated",
    2: "spoofing indicated",
    3: "multiple spoofing indications",
}
"""Spoof detection state from UBX-NAV-STATUS"""

PSMSTATUS = {
    0: "aquisition",
    1: "tracking",
    2: "power optimised tracking",
    3: "inactive",
}
"""Power save mode state from UBX-NAV-STATUS"""

# UBX-NAV-TIME*
UTCSTANDARD = {
    0: "Information not available",
    1: "Communications Research Labratory (CRL), Tokyo, Japan",
    2: "National Institute of Standards and Technology (NIST)",
    3: "U.S. Naval Observatory (USNO)",
    4: "International Bureau of Weights and Measures (BIPM)",
    5: "European laboratories",
    6: "Former Soviet Union (SU)",
    7: "National Time Service Center (NTSC), China",
    8: "National Physics Laboratory India (NPLI)",
    15: "Unknown",
}
"""UTC time standard from UBX-NAV-TIME*"""

SOURCEOFCURLS = {
    0: "Default",
    1: "Derived from time difference between GPS and GLONASS time",
    2: "GPS",
    3: "SBAS",
    4: "BeiDou",
    5: "Galileo",
    6: "Aided data",
    7: "Configured",
    8: "NavIC",
    255: "Unknown",
}
"""Information source for the current number of leap seconds, from UBX-NAV-TIMELS"""

SRCOFLSCHANGE = {
    0: "No source",
    2: "GPS",
    3: "SBAS",
    4: "BeiDou",
    5: "Galileo",
    6: "GLONASS",
    7: "NavIC",
}
"""Information source for the future leap second event, from UBX-NAV-TIMELS"""

ESFALG_AUTOMNTALGON = {
    0: ("AUTOOFF", "automatic alignment is not running"),
    1: ("AUTOON", "automatic alignment is running"),
}
"""autoMntAlgOn from ESF-ALG"""

ESFALG_STATUS = {
    0: ("FIXEDUSED", "user-defined/fixed angles are used"),
    1: ("ALIGNINGRP", "IMU-mount roll/pitch angles alignment is ongoing"),
    2: ("ALIGNINGRPW", "IMU-mount roll/pitch/yaw angles alignment is ongoing"),
    3: ("COARSEUSED", "coarse IMU-mount alignment are used"),
    4: ("FINEUSED", "fine IMU-mount alignment are used"),
}
"""status from ESF-ALG"""



"""
Collection of UBX helper methods which can be used
outside the UBXMessage or UBXReader classes.

Created on 15 Dec 2020

:author: semuadmin
:copyright: SEMU Consulting © 2020
:license: BSD 3-Clause
"""

import struct
from datetime import datetime, timedelta
from math import cos, pi, sin, trunc

from pynmeagps.nmeatypes_core import NMEA_HDR

import pyubx2.exceptions as ube
import pyubx2.ubxtypes_configdb as ubcdb
import pyubx2.ubxtypes_core as ubt
from pyubx2.ubxtypes_core import (
    ATTTYPE,
    NMEA_PROTOCOL,
    POLL,
    RTCM3_PROTOCOL,
    SET,
    UBX_HDR,
    UBX_PROTOCOL,
)
from pyubx2.ubxtypes_decodes import FIXTYPE, GNSSLIST

EPOCH0 = datetime(1980, 1, 6)  # EPOCH start date
LEAPOFFSET = 18  # leap year offset in seconds, valid as from 1/1/2017
SIW = 604800  # seconds in week = 3600*24*7


def att2idx(att: str) -> object:
    """
    Get integer indices corresponding to grouped attribute.

    e.g. svid_06 -> 6; gnssId_103 -> 103, gsid_03_04 -> (3,4), tow -> 0

    :param str att: grouped attribute name e.g. svid_01
    :return: indices as integer(s), or 0 if not grouped
    :rtype: int or tuple for nested group
    """

    try:
        att = att.split("_")
        ln = len(att)
        if ln == 2:  # one group level
            return int(att[1])
        if ln > 2:  # nested group level(s)
            return tuple(int(att[i]) for i in range(1, ln))
        return 0  # not grouped
    except ValueError:
        return 0


def att2name(att: str) -> str:
    """
    Get name of grouped attribute.

    e.g. svid_06 -> svid; gnssId_103 -> gnssId, tow -> tow

    :param str att: grouped attribute name e.g. svid_01
    :return: name without index e.g. svid
    :rtype: str
    """

    return att.split("_")[0]


def calc_checksum(content: bytes) -> bytes:
    """
    Calculate checksum using 8-bit Fletcher's algorithm.

    :param bytes content: message content, excluding header and checksum bytes
    :return: checksum
    :rtype: bytes

    """

    check_a = 0
    check_b = 0

    for char in content:
        check_a += char
        check_a &= 0xFF
        check_b += check_a
        check_b &= 0xFF

    return bytes((check_a, check_b))


def isvalid_checksum(message: bytes) -> bool:
    """
    Validate message checksum.

    :param bytes message: message including header and checksum bytes
    :return: checksum valid flag
    :rtype: bool

    """

    lenm = len(message)
    ckm = message[lenm - 2 : lenm]
    return ckm == calc_checksum(message[2 : lenm - 2])


def atttyp(att: str) -> str:
    """
    Helper function to return attribute type as string.

    :param str: attribute type e.g. 'U002'
    :return: type of attribute as string e.g. 'U'
    :rtype: str

    """

    return att[0:1]


def attsiz(att: str) -> int:
    """
    Helper function to return attribute size in bytes.

    :param str: attribute type e.g. 'U002'
    :return: size of attribute in bytes, or -1 if variable length
    :rtype: int

    """

    if att == "CH":  # variable length
        return -1
    return int(att[1:4])


def itow2utc(itow: int) -> datetime.time:
    """
    Convert GPS Time Of Week to UTC time

    :param int itow: GPS Time Of Week in milliseconds
    :return: UTC time hh.mm.ss
    :rtype: datetime.time

    """

    utc = EPOCH0 + timedelta(seconds=(itow / 1000) - LEAPOFFSET)
    return utc.time()


def utc2itow(utc: datetime) -> tuple:
    """
    Convert UTC datetime to GPS Week Number, Time Of Week

    :param datetime utc: datetime
    :return: GPS Week Number, Time of Week in milliseconds
    :rtype: tuple

    """

    wno = int((utc - EPOCH0).total_seconds() / SIW)
    sow = EPOCH0 + timedelta(seconds=wno * SIW)
    itow = int(((utc - sow).total_seconds() + LEAPOFFSET) * 1000)
    return wno, itow


def gpsfix2str(fix: int) -> str:
    """
    Convert GPS fix integer to descriptive string.

    :param int fix: GPS fix type (0-5)
    :return: GPS fix type as string
    :rtype: str

    """

    try:
        return FIXTYPE[fix]
    except KeyError:
        return str(fix)


def dop2str(dop: float) -> str:
    """
    Convert Dilution of Precision float to descriptive string.

    :param float dop: dilution of precision as float
    :return: dilution of precision as string
    :rtype: str

    """

    if dop == 0:
        dops = "N/A"
    elif dop <= 1:
        dops = "Ideal"
    elif dop <= 2:
        dops = "Excellent"
    elif dop <= 5:
        dops = "Good"
    elif dop <= 10:
        dops = "Moderate"
    elif dop <= 20:
        dops = "Fair"
    else:
        dops = "Poor"
    return dops


def gnss2str(gnss_id: int) -> str:
    """
    Convert GNSS ID to descriptive string
    ('GPS', 'GLONASS', etc.).

    :param int gnss_id: GNSS identifier as integer (0-6)
    :return: GNSS identifier as string
    :rtype: str

    """

    try:
        return GNSSLIST[gnss_id]
    except KeyError:
        return str(gnss_id)


def key_from_val(dictionary: dict, value) -> str:
    """
    Helper method - get dictionary key corresponding to (unique) value.

    :param dict dictionary: dictionary
    :param object value: unique dictionary value
    :return: dictionary key
    :rtype: str
    :raises: KeyError: if no key found for value

    """

    val = None
    for key, val in dictionary.items():
        if val == value:
            return key
    raise KeyError(f"No key found for value {value}")


def get_bits(bitfield: bytes, bitmask: int) -> int:
    """
    Get integer value of specified (masked) bit(s) in a UBX bitfield (attribute type 'X')

    e.g. to get value of bits 6,7 in bitfield b'\\\\x89' (binary 0b10001001)::

        get_bits(b'\\x89', 0b11000000) = get_bits(b'\\x89', 192) = 2

    :param bytes bitfield: bitfield byte(s)
    :param int bitmask: bitmask as integer (= Σ(2**n), where n is the number of the bit)
    :return: value of masked bit(s)
    :rtype: int
    """

    i = 0
    val = int(bitfield.hex(), 16)
    while bitmask & 1 == 0:
        bitmask = bitmask >> 1
        i += 1
    return val >> i & bitmask


def val2bytes(val, att: str) -> bytes:
    """
    Convert value to bytes for given UBX attribute type.

    :param object val: attribute value e.g. 25
    :param str att: attribute type e.g. 'U004'
    :return: attribute value as bytes
    :rtype: bytes
    :raises: UBXTypeError

    """

    try:
        if not isinstance(val, ATTTYPE[atttyp(att)]):
            raise TypeError(
                f"Attribute type {att} value {val} must be {ATTTYPE[atttyp(att)]}, not {type(val)}"
            )
    except KeyError as err:
        raise ube.UBXTypeError(f"Unknown attribute type {att}") from err

    if atttyp(att) == "X":  # byte
        valb = val
    elif atttyp(att) == "C":  # char
        valb = val.encode("utf-8", "backslashreplace") if isinstance(val, str) else val
    elif atttyp(att) in ("E", "I", "L", "U"):  # integer
        valb = val.to_bytes(attsiz(att), byteorder="little", signed=atttyp(att) == "I")
    elif atttyp(att) == "R":  # floating point
        valb = struct.pack("<f" if attsiz(att) == 4 else "<d", float(val))
    elif atttyp(att) == "A":  # array of unsigned integers
        valb = b""
        for i in range(attsiz(att)):
            valb += val[i].to_bytes(1, byteorder="little", signed=False)
    return valb


def bytes2val(valb: bytes, att: str) -> object:
    """
    Convert bytes to value for given UBX attribute type.

    :param bytes valb: attribute value in byte format e.g. b'\\\\x19\\\\x00\\\\x00\\\\x00'
    :param str att: attribute type e.g. 'U004'
    :return: attribute value as int, float, str or bytes
    :rtype: object
    :raises: UBXTypeError

    """

    if att == ubt.CH:  # single variable-length string (e.g. INF-NOTICE)
        val = valb.decode("utf-8", "backslashreplace")
    elif atttyp(att) in ("X", "C"):
        val = valb
    elif atttyp(att) in ("E", "I", "L", "U"):  # integer
        val = int.from_bytes(valb, byteorder="little", signed=atttyp(att) == "I")
    elif atttyp(att) == "R":  # floating point
        val = struct.unpack("<f" if attsiz(att) == 4 else "<d", valb)[0]
    elif atttyp(att) == "A":  # array of unsigned integers
        val = []
        for i in range(attsiz(att)):
            val.append(valb[i])
    else:
        raise ube.UBXTypeError(f"Unknown attribute type {att}")
    return val


def nomval(att: str) -> object:
    """
    Get nominal value for given UBX attribute type.

    :param str att: attribute type e.g. 'U004'
    :return: attribute value as int, float, str or bytes
    :rtype: object
    :raises: UBXTypeError

    """

    if att == "CH":
        val = ""
    elif atttyp(att) in ("X", "C"):
        val = b"\x00" * attsiz(att)
    elif atttyp(att) == "R":
        val = 0.0
    elif atttyp(att) in ("E", "I", "L", "U"):
        val = 0
    elif atttyp(att) == "A":  # array of unsigned integers
        val = [0] * attsiz(att)
    else:
        raise ube.UBXTypeError(f"Unknown attribute type {att}")
    return val


def msgclass2bytes(msgclass: int, msgid: int) -> bytes:
    """
    Convert message class/id integers to bytes.

    :param int msgClass: message class as integer e.g. 6
    :param int msgID: message ID as integer e.g. 1
    :return: message class as bytes e.g. b'/x06/x01'
    :rtype: bytes

    """

    return (val2bytes(msgclass, ubt.U1), val2bytes(msgid, ubt.U1))


def msgstr2bytes(msgclass: str, msgid: str) -> bytes:
    """
    Convert plain text UBX message class to bytes.

    :param str msgClass: message class as str e.g. 'CFG'
    :param str msgID: message ID as str e.g. 'CFG-MSG'
    :return: message class as bytes e.g. b'/x06/x01'
    :rtype: bytes
    :raises: UBXMessageError

    """

    try:
        return (
            key_from_val(ubt.UBX_CLASSES, msgclass),
            key_from_val(ubt.UBX_MSGIDS, msgid)[1:2],
        )
    except KeyError as err:
        raise ube.UBXMessageError(
            f"Undefined message, class {msgclass}, id {msgid}"
        ) from err


def cfgname2key(name: str) -> tuple:
    """
    Return hexadecimal key and data type for given
    configuration database key name.

    :param str name: config key as string e.g. "CFG_NMEA_PROTVER"
    :return: tuple of (key, type)
    :rtype: tuple: (int, str)
    :raises: UBXMessageError

    """
    try:
        return ubcdb.UBX_CONFIG_DATABASE[name]
    except KeyError as err:
        raise ube.UBXMessageError(
            f"Undefined configuration database key {name}"
        ) from err


def cfgkey2name(keyid: int) -> tuple:
    """
    Return key name and data type for given
    configuration database hexadecimal key.

    :param int keyID: config key as integer e.g. 0x20930001
    :return: tuple of (keyname, type)
    :rtype: tuple: (str, str)
    :raises: UBXMessageError

    """

    try:
        val = None
        for key, val in ubcdb.UBX_CONFIG_DATABASE.items():
            (kid, typ) = val
            if keyid == kid:
                return (key, typ)

        # undocumented configuration database key
        # type is derived from keyID
        key = f"CFG_{hex(keyid)}"
        typ = f"X{ubcdb.UBX_CONFIG_STORSIZE[int(hex(keyid)[2:3])]:03d}"
        return (key, typ)

    except KeyError as err:
        raise ube.UBXMessageError(
            f"Invalid configuration database key {hex(keyid)}"
        ) from err


def protocol(raw: bytes) -> int:
    """
    Gets protocol of raw message.

    :param bytes raw: raw (binary) message
    :return: protocol type (1 = NMEA, 2 = UBX, 4 = RTCM3, 0 = unknown)
    :rtype: int
    """

    p = raw[0:2]
    if p == UBX_HDR:
        return UBX_PROTOCOL
    if p in NMEA_HDR:
        return NMEA_PROTOCOL
    if p[0] == 0xD3 and (p[1] & ~0x03) == 0:
        return RTCM3_PROTOCOL
    return 0


def hextable(raw: bytes, cols: int = 8) -> str:
    """
    Formats raw (binary) message in tabular hexadecimal format e.g.

    000: 2447 4e47 5341 2c41 2c33 2c33 342c 3233 | b'$GNGSA,A,3,34,23' |

    :param bytes raw: raw (binary) data
    :param int cols: number of columns in hex table (8)
    :return: table of hex data
    :rtype: str
    """

    hextbl = ""
    colw = cols * 4
    rawh = raw.hex()
    for i in range(0, len(rawh), colw):
        rawl = rawh[i : i + colw].ljust(colw, " ")
        hextbl += f"{int(i/2):03}: "
        for col in range(0, colw, 4):
            hextbl += f"{rawl[col : col + 4]} "
        hextbl += f" | {bytes.fromhex(rawl)} |\n"

    return hextbl


def cel2cart(elevation: float, azimuth: float) -> tuple:
    """
    Convert celestial coordinates (degrees) to Cartesian coordinates.

    :param float elevation: elevation
    :param float azimuth: azimuth
    :return: cartesian x,y coordinates
    :rtype: tuple
    """

    if not (isinstance(elevation, (float, int)) and isinstance(azimuth, (float, int))):
        return (0, 0)
    ele, azi = [c * pi / 180 for c in (elevation, azimuth)]
    x = cos(azi) * cos(ele)
    y = sin(azi) * cos(ele)
    return (x, y)


def escapeall(val: bytes) -> str:
    """
    Escape all byte characters e.g. b'\\\\x73' rather than b`s`

    :param bytes val: bytes
    :return: string of escaped bytes
    :rtype: str
    """

    return "b'{}'".format("".join(f"\\x{b:02x}" for b in val))


def val2sphp(val: float, scale: float = 1e-7) -> tuple:
    """
    Convert a float value into separate standard and high precisions components,
    multiplied by a scaling factor to render them as integers, as required by some
    CFG and NAV messages.

    e.g. 48.123456789 becomes (481234567, 89)

    :param float val: value as float
    :param float scale: scaling factor e.g. 1e-7
    :return: tuple of (standard precision, high precision)
    :rtype: tuple
    """

    val = val / scale
    val_sp = trunc(val)
    val_hp = round((val - val_sp) * 100)
    return val_sp, val_hp


def getinputmode(data: bytes) -> int:
    """
    Return input message mode (SET or POLL).

    :param bytes data: raw UBX input message
    :return: message mode (1 = SET, 2 = POLL)
    :rtype: int
    """

    if (
        len(data) == 8
        or data[2:4] == b"\x06\x8b"  # CFG-VALGET
        or (
            data[2:4]
            in (
                b"\x06\x01",
                b"\x06\x02",
                b"\x06\x03",
                b"\x06\x31",
            )  # CFG-INF, CFG-MSG, CFG-PRT, CFG-TP5
            and len(data) <= 10
        )
    ):
        return POLL
    return SET


def process_monver(msg: object) -> dict:
    """
    Process parsed MON-VER sentence into dictionary of
    hardware, firmware and software version identifiers.

    :param UBXMessage msg: UBX MON-VER config message
    :return: dict of version information
    :rtype: dict
    """

    exts = []
    fw_version = "N/A"
    rom_version = "N/A"
    gnss_supported = ""
    model = ""
    sw_version = getattr(msg, "swVersion", b"N/A")
    sw_version = sw_version.replace(b"\x00", b"").decode()
    sw_version = sw_version.replace("ROM CORE", "ROM")
    sw_version = sw_version.replace("EXT CORE", "Flash")
    hw_version = getattr(msg, "hwVersion", b"N/A")
    hw_version = hw_version.replace(b"\x00", b"").decode()

    for i in range(9):
        ext = getattr(msg, f"extension_{i+1:02d}", b"")
        ext = ext.replace(b"\x00", b"").decode()
        exts.append(ext)
        if "FWVER=" in exts[i]:
            fw_version = exts[i].replace("FWVER=", "")
        if "PROTVER=" in exts[i]:
            rom_version = exts[i].replace("PROTVER=", "")
        if "PROTVER " in exts[i]:
            rom_version = exts[i].replace("PROTVER ", "")
        if "MOD=" in exts[i]:
            model = exts[i].replace("MOD=", "")
            hw_version = f"{model} {hw_version}"
        for gnss in (
            "GPS",
            "GLO",
            "GAL",
            "BDS",
            "SBAS",
            "IMES",
            "QZSS",
            "NAVIC",
        ):
            if gnss in exts[i]:
                gnss_supported = gnss_supported + gnss + " "

    verdata = {}
    verdata["swversion"] = sw_version
    verdata["hwversion"] = hw_version
    verdata["fwversion"] = fw_version
    verdata["romversion"] = rom_version
    verdata["gnss"] = gnss_supported

    return verdata


def val2twoscomp(val: int, att: str) -> int:
    """
    Convert signed integer to two's complement binary representation.

    :param int val: value
    :param str att: attribute type e.g. "U024"
    :return: two's complement representation of value
    :rtype: int
    """

    return val & pow(2, attsiz(att)) - 1


def val2signmag(val: int, att: str) -> int:
    """
    Convert signed integer to sign magnitude binary representation.

    High-order bit represents sign (0 +ve, 1 -ve).

    :param int val: value
    :param str att: attribute type e.g. "U024"
    :return: sign magnitude representation of value
    :rtype: int
    """

    return (abs(val) & pow(2, attsiz(att)) - 1) | ((1 if val < 0 else 0) << attsiz(att))



"""
test pyubx2.ubxtypes_enums.py

Created on 3 Oct 2020

*** NB: must be saved in UTF-8 format ***

@author: semuadmin
"""

import os
import unittest

from pyubx2.ubxtypes_decodes import *


class StaticTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        dirname = os.path.dirname(__file__)

    def tearDown(self):
        pass

    def testenums(self):
        for enum in (
            DGNSMODE,
            CONFLVL,
            SIGCFMASK,
            PROTOCOLID,
            DYNMODEL,
            FIXMODE,
            NMEAVERSION,
            SVNUMBERING,
            MAINTALKERID,
            GSVTALKERID,
            ODOPROFILE,
            CHARLEN,
            PARITY,
            NSTOPBITS,
            POL,
            SPIMODE,
            STATE,
            TIMEREF,
            NAVBBRMASK,
            RESETMODE,
            MODE,
            GRIDUTCGNSS,
            PROTIDS,
            ASTATUS,
            APOWER,
            JAMMINGSTATE,
            BOOTTYPE,
            GEOFENCE_STATUS,
            COMBSTATE,
            HEALTH,
            VISIBILITY,
            PLPOSFRAME,
            PLVELFRAME,
            GPSFIX,
            FIXTYPE,
            PSMSTATE,
            CARRSOLN,
            QUALITYIND,
            ORBITSOURCE,
            SBASMODE,
            SBASSYS,
            SBASINTEGRITYUSED,
            CORRSOURCE,
            IONOMODEL,
            SIGID,
            SOURCEOFCURLS,
            SRCOFLSCHANGE,
            SPOOFDETSTATE,
            PSMSTATUS,
            UTCSTANDARD,
        ):
            for key, val in enum.items():
                # print(key, val)
                pass  # test is just for test completion stats


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
Constructor method tests for pyubx2.UBXMessage

Created on 21 Oct 2020

*** NB: must be saved in UTF-8 format ***

@author: semuadmin
"""
# pylint: disable=line-too-long, invalid-name, missing-docstring, no-member

import unittest

from pyubx2 import UBXMessage, UBXReader, GET, SET, POLL


class FillTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def tearDown(self):
        pass

    def testFill_CFGMSG(self):  # test POLL constructor fill, format 1
        EXPECTED_RESULT = "<UBX(CFG-MSG, msgClass=NMEA-Standard, msgID=VTG)>"
        res = UBXMessage("CFG", "CFG-MSG", POLL, msgClass=240, msgID=5)
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_CFGMSG2(self):  # test POLL constructor fill, format 2
        EXPECTED_RESULT = "<UBX(CFG-MSG, msgClass=NMEA-Standard, msgID=VTG)>"
        res = UBXMessage(b"\x06", b"\x01", POLL, msgClass=240, msgID=5)
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_CFGMSG3(self):  # test POLL constructor fill, format 3
        EXPECTED_RESULT = "<UBX(CFG-MSG, msgClass=NMEA-Standard, msgID=VTG)>"
        res = UBXMessage(6, 1, POLL, msgClass=240, msgID=5)
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_CFGMSG4(self):  # test SET constructor fill
        EXPECTED_RESULT = "<UBX(CFG-MSG, msgClass=NMEA-Standard, msgID=GLL, rateDDC=0, rateUART1=1, rateUART2=0, rateUSB=1, rateSPI=0, reserved=0)>"
        res = UBXMessage(
            "CFG", "CFG-MSG", SET, msgClass=240, msgID=1, rateUART1=1, rateUSB=1
        )
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_INFNOTICE(
        self,
    ):  # test INF-NOTICE variable length message constructor fill
        EXPECTED_RESULT = "<UBX(INF-NOTICE, message=Lorem ipsum dolor sit amet)>"
        res = UBXMessage("INF", "INF-NOTICE", GET, message="Lorem ipsum dolor sit amet")
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_CFGNMEA(self):  # test SET constructor fill, set all values
        EXPECTED_RESULT = "<UBX(CFG-NMEA, posFilt=0, mskPosFilt=0, timeFilt=0, dateFilt=0, gpsOnlyFilter=0, trackFilt=0, nmeaVersion=64, numSV=4, compat=0, consider=0, limit82=0, highPrec=0, gps=0, sbas=0, galileo=0, qzss=0, glonass=0, beidou=0, svNumbering=0, mainTalkerId=0, gsvTalkerId=0, version=0, bdsTalkerId=b'\\x00\\x00', reserved1=0)>"
        res = UBXMessage(
            "CFG",
            "CFG-NMEA",
            SET,
            filter=b"\x45",
            nmeaVersion=64,
            numSV=4,
            flags=b"\x14",
        )
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_CFGNMEA2(
        self,
    ):  # test SET constructor fill, set some values, default others
        EXPECTED_RESULT = "<UBX(CFG-NMEA, posFilt=0, mskPosFilt=0, timeFilt=0, dateFilt=0, gpsOnlyFilter=0, trackFilt=0, nmeaVersion=35, numSV=1, compat=0, consider=0, limit82=0, highPrec=0, gps=0, sbas=0, galileo=0, qzss=0, glonass=0, beidou=0, svNumbering=0, mainTalkerId=0, gsvTalkerId=0, version=0, bdsTalkerId=b'\\x00\\x00', reserved1=0)>"
        res = UBXMessage("CFG", "CFG-NMEA", SET, nmeaVersion=35, numSV=1)
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_CFGNMEAPARSE(
        self,
    ):  # check that raw payload is correctly populated and parses back to original message
        EXPECTED_RESULT = "<UBX(CFG-NMEA, posFilt=0, mskPosFilt=0, timeFilt=0, dateFilt=0, gpsOnlyFilter=0, trackFilt=0, nmeaVersion=35, numSV=1, compat=0, consider=0, limit82=0, highPrec=0, gps=0, sbas=0, galileo=0, qzss=0, glonass=0, beidou=0, svNumbering=0, mainTalkerId=0, gsvTalkerId=0, version=0, bdsTalkerId=b'\\x00\\x00', reserved1=0)>"
        res = UBXMessage("CFG", "CFG-NMEA", SET, nmeaVersion=35, numSV=1)
        res2 = UBXReader.parse(res.serialize())
        self.assertEqual(str(res2), EXPECTED_RESULT)

    def testFill_CFGNMEAPOLL(self):  # test POLL constructor, no payload
        EXPECTED_RESULT = "<UBX(CFG-NMEA)>"
        res = UBXMessage("CFG", "CFG-NMEA", POLL)
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_CFGNMEAPOLL2(self):  # test POLL constructor, no payload
        EXPECTED_RESULT = "<UBX(CFG-NMEA)>"
        res = UBXMessage("CFG", "CFG-NMEA", POLL)
        res2 = UBXReader.parse(res.serialize())
        self.assertEqual(str(res2), EXPECTED_RESULT)

    def testFill_CFGGNSS(self):  #  test CFG-GNSS SET multiple repeats in group
        EXPECTED_RESULT = "<UBX(CFG-GNSS, msgVer=0, numTrkChHw=2, numTrkChUse=4, numConfigBlocks=2, gnssId_01=GPS, resTrkCh_01=4, maxTrkCh_01=32, reserved0_01=0, enable_01=1, sigCfMask_01=4, gnssId_02=GLONASS, resTrkCh_02=3, maxTrkCh_02=24, reserved0_02=0, enable_02=0, sigCfMask_02=64)>"
        res = UBXMessage(
            "CFG",
            "CFG-GNSS",
            SET,
            numTrkChHw=2,
            numTrkChUse=4,
            numConfigBlocks=2,
            gnssId_01=0,
            resTrkCh_01=4,
            maxTrkCh_01=32,
            enable_01=1,
            sigCfMask_01=4,
            gnssId_02=6,
            resTrkCh_02=3,
            maxTrkCh_02=24,
            enable_02=0,
            sigCfMask_02=64,
        )
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_CFGGNSS_NOBITFIELD(
        self,
    ):  #  test CFG-GNSS SET with parsebitfield = False
        EXPECTED_RESULT = "<UBX(CFG-GNSS, msgVer=0, numTrkChHw=2, numTrkChUse=4, numConfigBlocks=2, gnssId_01=GPS, resTrkCh_01=4, maxTrkCh_01=32, reserved0_01=0, flags_01=b'\\x01\\x00\\x04\\x00', gnssId_02=GLONASS, resTrkCh_02=3, maxTrkCh_02=24, reserved0_02=0, flags_02=b'\\x00\\x00\\x40\\x00')>"
        EXPECTED_RESULT2 = "<UBX(CFG-GNSS, msgVer=0, numTrkChHw=2, numTrkChUse=4, numConfigBlocks=2, gnssId_01=GPS, resTrkCh_01=4, maxTrkCh_01=32, reserved0_01=0, enable_01=1, sigCfMask_01=4, gnssId_02=GLONASS, resTrkCh_02=3, maxTrkCh_02=24, reserved0_02=0, enable_02=0, sigCfMask_02=64)>"
        res = UBXMessage(
            "CFG",
            "CFG-GNSS",
            SET,
            parsebitfield=False,
            numTrkChHw=2,
            numTrkChUse=4,
            numConfigBlocks=2,
            gnssId_01=0,
            resTrkCh_01=4,
            maxTrkCh_01=32,
            flags_01=b"\x01\x00\x04\x00",
            gnssId_02=6,
            resTrkCh_02=3,
            maxTrkCh_02=24,
            flags_02=b"\x00\x00\x40\x00",
        )
        self.assertEqual(str(res), EXPECTED_RESULT)
        res2 = UBXReader.parse(
            res.serialize()
        )  # reconstruct message and parse again with parsebitfield = True
        self.assertEqual(str(res2), EXPECTED_RESULT2)

    def testFill_CFGDOSC(self):  # test CFG-DOSC multiple repeats in group
        EXPECTED_RESULT = "<UBX(CFG-DOSC, version=23, numOsc=2, reserved1=0, oscId_01=4, reserved2_01=0, isCalibrated_01=0, controlIf_01=0, freq_01=22, phaseOffset_01=0, withTemp_01=0, withAge_01=0, timeToTemp_01=0, reserved3_01=0, gainVco_01=0, gainUncertainty_01=0, reserved4_01=0, oscId_02=7, reserved2_02=0, isCalibrated_02=0, controlIf_02=0, freq_02=44, phaseOffset_02=0, withTemp_02=0, withAge_02=0, timeToTemp_02=0, reserved3_02=0, gainVco_02=0, gainUncertainty_02=0, reserved4_02=0)>"
        res = UBXMessage(
            "CFG",
            "CFG-DOSC",
            SET,
            version=23,
            numOsc=2,
            oscId_01=4,
            freq_01=22,
            oscId_02=7,
            freq_02=44,
        )
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_CFGDOSC1(self):  # test CFG-DOSC single repeat in group
        EXPECTED_RESULT = "<UBX(CFG-DOSC, version=37, numOsc=1, reserved1=0, oscId_01=8, reserved2_01=0, isCalibrated_01=0, controlIf_01=0, freq_01=53, phaseOffset_01=26, withTemp_01=0, withAge_01=0, timeToTemp_01=0, reserved3_01=0, gainVco_01=4, gainUncertainty_01=0.3, reserved4_01=0)>"
        res = UBXMessage(
            "CFG",
            "CFG-DOSC",
            SET,
            version=37,
            numOsc=1,
            oscId_01=8,
            freq_01=53,
            phaseOffset_01=26,
            gainVco_01=4,
            gainUncertainty_01=0.3,
        )
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_CFGDOSCPARSE(
        self,
    ):  # test CFG-DOSC check that raw payload is correctly populated and parses back to original message
        EXPECTED_RESULT = "<UBX(CFG-DOSC, version=37, numOsc=1, reserved1=0, oscId_01=8, reserved2_01=0, isCalibrated_01=0, controlIf_01=0, freq_01=53.0, phaseOffset_01=26, withTemp_01=0.0, withAge_01=0.0, timeToTemp_01=0, reserved3_01=0, gainVco_01=4.0, gainUncertainty_01=0.296875, reserved4_01=0)>"
        res = UBXMessage(
            "CFG",
            "CFG-DOSC",
            SET,
            version=37,
            numOsc=1,
            oscId_01=8,
            freq_01=53,
            phaseOffset_01=26,
            gainVco_01=4,
            gainUncertainty_01=0.3,
        )
        res2 = UBXReader.parse(res.serialize())
        self.assertEqual(str(res2), EXPECTED_RESULT)

    def testFill_CFGDOSC2(self):  # test CFG-DOSC empty group
        EXPECTED_RESULT = "<UBX(CFG-DOSC, version=37, numOsc=0, reserved1=0)>"
        res = UBXMessage("CFG", "CFG-DOSC", SET, version=37, numOsc=0)
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_CFGDAT(
        self,
    ):  # test CFG-DAT floating point attribute, single and double precision
        EXPECTED_RESULT = "<UBX(CFG-DAT, majA=4321.123456789128, flat=-2964.00469836, dX=-1.2345678, dY=27.40654, dZ=0.0, rotX=0.0, rotY=0.0, rotZ=0.0, scale=0.0)>"
        res = UBXMessage(
            "CFG",
            "CFG-DAT",
            SET,
            majA=4321.123456789128,
            flat=-2964.00469836,
            dX=-1.2345678,
            dY=27.40654,
        )
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_CFGDATPARSE(
        self,
    ):  # check that raw payload is correctly populated and parses back to original message
        EXPECTED_RESULT = "<UBX(CFG-DAT, datumNum=0, datumName=b'\\x00\\x00\\x00\\x00\\x00\\x00', majA=4321.123456789128, flat=-2964.00469836, dX=-1.2345677614212036, dY=27.406539916992188, dZ=0.0, rotX=0.0, rotY=0.0, rotZ=0.0, scale=0.0)>"
        res = UBXMessage(
            "CFG",
            "CFG-DAT",
            GET,
            majA=4321.123456789128,
            flat=-2964.00469836,
            dX=-1.2345678,
            dY=27.40654,
        )
        res2 = UBXReader.parse(res.serialize(), msgmode=GET)
        self.assertEqual(str(res2), EXPECTED_RESULT)

    def testFill_CFGDATPARSE2(
        self,
    ):  # check that raw payload is correctly populated and parses back to original message
        EXPECTED_RESULT = "<UBX(CFG-DAT, majA=0.0, flat=0.0, dX=-1.2345677614212036, dY=27.406539916992188, dZ=0.0, rotX=0.0, rotY=0.0, rotZ=0.0, scale=0.0)>"
        res = UBXMessage(
            "CFG",
            "CFG-DAT",
            SET,
            dX=-1.2345678,
            dY=27.40654,
        )
        res2 = UBXReader.parse(res.serialize(), msgmode=SET)
        self.assertEqual(str(res2), EXPECTED_RESULT)

    def testFill_MONSPAN(self):  # test MON-SPAN with nominal spectrum array values
        EXPECTED_RESULT = "<UBX(MON-SPAN, version=1, numRfBlocks=2, reserved0=0, spectrum_01=[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], span_01=0, res_01=0, center_01=0, pga_01=0, reserved1_01=0, spectrum_02=[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], span_02=0, res_02=0, center_02=0, pga_02=0, reserved1_02=0)>"
        res = UBXMessage(
            "MON",
            "MON-SPAN",
            GET,
            version=1,
            numRfBlocks=2,
        )
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testFill_MONSPAN2(self):  # test MON-SPAN with specified spectrum array values
        EXPECTED_RESULT = "<UBX(MON-SPAN, version=2, numRfBlocks=1, reserved0=0, spectrum_01=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255], span_01=0, res_01=0, center_01=0, pga_01=0, reserved1_01=0)>"
        spec = []
        for i in range(256):
            spec.append(i)
        res = UBXMessage(
            "MON",
            "MON-SPAN",
            GET,
            version=2,
            numRfBlocks=1,
            spectrum_01=spec,
        )
        res2 = UBXMessage(
            "MON",
            "MON-SPAN",
            GET,
            payload=res.payload,
        )
        self.assertEqual(str(res), EXPECTED_RESULT)
        self.assertEqual(str(res2), EXPECTED_RESULT)

    def testEVAL(
        self,
    ):  # double check that eval of repr(msg) reproduces original message
        res1 = UBXMessage("CFG", "CFG-MSG", POLL, msgClass=240, msgID=5)
        res2 = eval(repr(res1))
        self.assertEqual(str(res1), str(res2))

    def testEVAL2(
        self,
    ):  # double check that eval of repr(msg) reproduces original message
        res1 = UBXMessage(
            "CFG", "CFG-MSG", SET, msgClass=240, msgID=5, rateUART1=1, rateUSB=1
        )
        res2 = eval(repr(res1))
        self.assertEqual(str(res1), str(res2))

    def testFill_NAVSBAS(self):  #  test NAV-SBAS GET constructor with bit flags
        EXPECTED_RESULT = "<UBX(NAV-SBAS, iTOW=23:59:42, geo=1, mode=0, sys=0, Ranging=1, Corrections=0, Integrity=1, Testmode=0, Bad=1, cnt=0, integrityUsed=1, reserved1=0)>"
        res = UBXMessage(
            "NAV",
            "NAV-SBAS",
            GET,
            geo=1,
            mode=0,
            sys=0,
            Ranging=1,
            Corrections=0,
            Integrity=1,
            Testmode=0,
            Bad=1,
            cnt=0,
            integrityUsed=1,
        )
        self.assertEqual(str(res), EXPECTED_RESULT)

    def testEval_NAVBAS(self):  # test payload is correctly set for bit flags
        res1 = UBXMessage(
            "NAV",
            "NAV-SBAS",
            GET,
            geo=1,
            mode=0,
            sys=0,
            Ranging=1,
            Corrections=0,
            Integrity=1,
            Testmode=0,
            Bad=1,
            numCh=0,
            integrityUsed=1,
        )
        res2 = UBXMessage("NAV", "NAV-SBAS", GET, payload=res1.payload)
        self.assertEqual(str(res1), str(res2))

    def testCFG_TP5(self):  # test CFG-TP5 constructor
        res1 = UBXMessage("CFG", "CFG-TP5", POLL)
        res2 = "<UBX(CFG-TP5)>"
        self.assertEqual(str(res1), res2)

    def testCFG_TP5_TPX(self):  # test CFG-TP5-TPX constructor
        res1 = UBXMessage("CFG", "CFG-TP5", POLL, payload=b"\x01")
        res2 = "<UBX(CFG-TP5, tpIdx=1)>"
        self.assertEqual(str(res1), res2)
        res1 = UBXMessage("CFG", "CFG-TP5", POLL, payload=b"\x00")
        res2 = "<UBX(CFG-TP5, tpIdx=0)>"
        self.assertEqual(str(res1), res2)
        res1 = UBXMessage("CFG", "CFG-TP5", POLL, tpIdx=0)
        res2 = "<UBX(CFG-TP5, tpIdx=0)>"
        self.assertEqual(str(res1), res2)
        res1 = UBXMessage("CFG", "CFG-TP5", POLL, tpIdx=1)
        res2 = "<UBX(CFG-TP5, tpIdx=1)>"
        self.assertEqual(str(res1), res2)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
Parse method tests for pyubx2.UBXMessage

Created on 3 Oct 2020

*** NB: must be saved in UTF-8 format ***

@author: semuadmin
"""
# pylint: disable=line-too-long, invalid-name, missing-docstring, no-member

import unittest

from pyubx2 import UBXMessage, UBXReader, VALCKSUM, VALNONE, SET, SETPOLL


class ParseTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.ack_ack = b"\xb5b\x05\x01\x02\x00\x06\x01\x0f\x38"
        self.ack_ack_badck = b"\xb5b\x05\x01\x02\x00\x06\x01\x0f\x37"
        self.cfg_msg = b"\xb5b\x06\x01\x08\x00\xf0\x01\x00\x01\x01\x01\x00\x00\x036"
        self.cfg_prt = b"\xb5b\x06\x00\x00\x00\x06\x18"
        self.nav_pl = b"\xb5b\x01\x62\x34\x00\x01\x05\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x48\x04\xA9\x14\x6D\x00\x00\x00\x6D\x00\x00\x00\x6D\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xEF\x88"
        self.nav_velned = b"\xb5b\x01\x12$\x000D\n\x18\xfd\xff\xff\xff\xf1\xff\xff\xff\xfc\xff\xff\xff\x10\x00\x00\x00\x0f\x00\x00\x00\x83\xf5\x01\x00A\x00\x00\x00\xf0\xdfz\x00\xd0\xa6"
        self.nav_svinfo = b""
        self.cfg_nmeavx = b"\xb5b\x06\x17\x04\x00\x00\x00\x00\x00\x21\xe9"
        self.cfg_nmeav0 = b"\xb5b\x06\x17\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x29\x61"
        self.mga_dbd = b"\xb5b\x13\x80\x0e\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x01\x02\xf2\xc2"
        self.mga_flash_ack = b"\xb5b\x13\x21\x06\x00\x03\x01\x02\x00\x00\x04\x44\x3a"
        self.cfg_valget = b"\xb5b\x06\x8b\x0c\x00\x00\x00\x00\x00\x01\x00\x52\x40\x80\x25\x00\x00\xd5\xd0"
        self.cfg_valget2 = (
            b"\xb5b\x06\x8b\x09\x00\x00\x00\x00\x00\x01\x00\x51\x20\x55\x61\xc2"
        )
        self.cfg_valget3 = b"\xb5b\x06\x8b\x16\x00\x00\x00\x00\x00\x01\x00\x51\x20\x55\x01\x00\x52\x40\x80\x25\x00\x00\x02\x00\x21\x30\x23\x1c\x92"
        self.cfg_valget4 = b"\xb5b\x06\x8b\x0c\x00\x00\x00\x00\x00\x68\x00\x11\x40\xb6\xf3\x9d\x3f\xdb\x3d"
        self.esf_meas = b"\xb5b\x10\x02\x1c\x00\x6d\xd8\x07\x00\x18\x20\x00\x00\xcd\x06\x00\x0e\xe4\xfe\xff\x0d\x03\xfa\xff\x05\x09\x0b\x00\x0c\x6d\xd8\x07\x00\xee\x51"
        self.esf_meas2 = b"\xb5b\x10\x02\x18\x00\x72\xd8\x07\x00\x18\x18\x00\x00\x4b\xfd\xff\x10\x40\x02\x00\x11\x23\x28\x00\x12\x72\xd8\x07\x00\x03\x9c"
        self.mga_ini1 = b"\xb5b\x13\x40\x14\x00\x01\x00\x01\x02\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x93\xc8"
        self.mon_span = b"\xb5b\n1\x14\x01\x00\x01\x00\x00-+-,+-.,-.+,+.-..-,..//./00203017?9398:L]<@C;H<>=A@BDCGJNQRVY[_cgpqyz\x7f\x84\x8c\x90\x99\xa0\xa7\xae\xb0\xae\xaa\xa7\xa2\x9b\x97\x96\x94\x91\x90\x8e\x8c\x8c\x8c\x8b\x8b\x89\x88\x89\x89\x89\x8b\x88\x89\x8a\x89\x8a\x8a\x89\x8a\x8b\x8a\x8a\x8b\x8b\x8c\x8a\x8a\x8a\x8b\x88\x88\x87\x87\x86\x85\x85\x85\x84\x89\x84\x85\x83\x84\x84\x84\x85\x88\x87\x87\x88\x8a\x8a\x8a\x8a\x8b\x8e\x8c\x8d\x8d\x8f\x8e\x8d\x8f\x8e\x8f\x8f\x8e\x8f\x8f\x90\x91\x92\x93\x93\x93\x95\x94\x94\x94\x94\x95\x94\x95\x93\x93\x91\x92\x93\x92\x94\x95\x94\x95\x97\x97\x98\x97\x94\x90\x8d\x86\x82\x7fyupmg`]VRLEB?=;99665422202101///-//.-0-.-/..--,.-+-,--+.,,--,,-*\x00 \xa1\x07 \xa1\x07\x00@\xc4`^\x0c\x00\x00\x00\x15j"
        self.tim_tp = b"\xb5b\r\x01\x10\x00\x88gh\x16\x00\x00\x00\x00\x00\x00\x00\x00\x85\x08\x1b\x0fB\xff"
        self.esf_status = b"\xb5b\x10\x10.\x00\x00\x00\x00\x00\x02$\x02\x00\x00\x00\x00\x00\x00\x00\x00\x06\x85\x04\x00\x00\r\x04\n\x00\x8e\x04\n\x00\x00\x04\n\x00\x01\x00\n\x00\x92\x04\x00\x00\x00\x00\x00\x00\x00\x00k4"

    def tearDown(self):
        pass

    def testAck(self):
        res = UBXReader.parse(self.ack_ack, validate=VALCKSUM)
        self.assertIsInstance(res, UBXMessage)

    def testAckID(self):
        res = UBXReader.parse(self.ack_ack)
        self.assertEqual(res.identity, "ACK-ACK")

    def testAckStr(self):
        res = UBXReader.parse(self.ack_ack, validate=VALCKSUM)
        self.assertEqual(str(res), "<UBX(ACK-ACK, clsID=CFG, msgID=CFG-MSG)>")

    def testAckRepr(self):
        res = UBXReader.parse(self.ack_ack)
        self.assertEqual(
            repr(res), "UBXMessage(b'\\x05', b'\\x01', 0, payload=b'\\x06\\x01')"
        )

    def testAckCkF(self):
        UBXReader.parse(self.ack_ack_badck, validate=VALNONE)

    def testCfg(self):
        res = UBXReader.parse(self.ack_ack)
        self.assertIsInstance(res, UBXMessage)

    def testCfgID(self):
        res = UBXReader.parse(self.cfg_msg, validate=VALCKSUM)
        self.assertEqual(res.identity, "CFG-MSG")

    def testCfgStr(self):
        res = UBXReader.parse(self.cfg_msg, validate=VALCKSUM)
        self.assertEqual(
            str(res),
            "<UBX(CFG-MSG, msgClass=NMEA-Standard, msgID=GLL, rateDDC=0, rateUART1=1, rateUART2=1, rateUSB=1, rateSPI=0, reserved=0)>",
        )

    def testCfgRepr(self):
        res = UBXReader.parse(self.cfg_msg)
        self.assertEqual(
            repr(res),
            "UBXMessage(b'\\x06', b'\\x01', 0, payload=b'\\xf0\\x01\\x00\\x01\\x01\\x01\\x00\\x00')",
        )

    def testCfgProp1(self):
        res = UBXReader.parse(self.cfg_msg, validate=VALCKSUM)
        self.assertEqual(res.rateUART1, 1)

    def testCfgProp2(self):
        res = UBXReader.parse(self.cfg_msg)
        self.assertEqual(res.rateSPI, 0)

    def testNavPL(self):
        res = UBXReader.parse(self.nav_pl, validate=VALCKSUM)
        self.assertEqual(res.plPos2, 109)

    def testNavVelNed(self):
        res = UBXReader.parse(self.nav_velned, validate=VALCKSUM)
        self.assertIsInstance(res, UBXMessage)

    def testNavVelNedID(self):
        res = UBXReader.parse(self.nav_velned)
        self.assertEqual(res.identity, "NAV-VELNED")

    def testNavVelNedStr(self):
        res = UBXReader.parse(self.nav_velned)
        self.assertEqual(
            str(res),
            "<UBX(NAV-VELNED, iTOW=16:01:48, velN=-3, velE=-15, velD=-4, speed=16, gSpeed=15, heading=1.28387, sAcc=65, cAcc=80.5272)>",
        )

    def testNavVelNedRepr(self):
        res = UBXReader.parse(self.nav_velned)
        self.assertEqual(
            repr(res),
            "UBXMessage(b'\\x01', b'\\x12', 0, payload=b'0D\\n\\x18\\xfd\\xff\\xff\\xff\\xf1\\xff\\xff\\xff\\xfc\\xff\\xff\\xff\\x10\\x00\\x00\\x00\\x0f\\x00\\x00\\x00\\x83\\xf5\\x01\\x00A\\x00\\x00\\x00\\xf0\\xdfz\\x00')",
        )

    def testNavVelNedProp1(self):
        res = UBXReader.parse(self.nav_velned, validate=VALCKSUM)
        self.assertEqual(res.iTOW, 403326000)

    def testNavVelNedProp2(self):
        res = UBXReader.parse(self.nav_velned)
        self.assertEqual(res.cAcc, 80.5272)

    def testCfgPrt(self):  # POLL example with null payload
        res = UBXReader.parse(self.cfg_prt)
        self.assertIsInstance(res, UBXMessage)

    def testCfgPrtID(self):
        res = UBXReader.parse(self.cfg_prt)
        self.assertEqual(res.identity, "CFG-PRT")

    def testCfgPrtStr(self):
        res = UBXReader.parse(self.cfg_prt, validate=VALCKSUM)
        self.assertEqual(str(res), "<UBX(CFG-PRT)>")

    def testCfgPrtRepr(self):
        res = UBXReader.parse(self.cfg_prt)
        self.assertEqual(repr(res), "UBXMessage(b'\\x06', b'\\x00', 0)")

    def testCfgNmeaVx(self):  # test older NMEA message parse
        res = UBXReader.parse(self.cfg_nmeavx)
        self.assertEqual(
            str(res),
            "<UBX(CFG-NMEA, posFilt=0, mskPosFilt=0, timeFilt=0, dateFilt=0, gpsOnlyFilter=0, trackFilt=0, nmeaVersion=0, numSV=0, compat=0, consider=0, limit82=0, highPrec=0)>",
        )

    def testCfgNmeaV0(self):  # test older NMEA message parse
        res = UBXReader.parse(self.cfg_nmeav0)
        self.assertEqual(
            str(res),
            "<UBX(CFG-NMEA, posFilt=0, mskPosFilt=0, timeFilt=0, dateFilt=0, gpsOnlyFilter=0, trackFilt=0, nmeaVersion=0, numSV=0, compat=0, consider=0, limit82=0, highPrec=0, gps=0, sbas=0, galileo=0, qzss=0, glonass=0, bBeidou=0, svNumbering=0, mainTalkerId=0, gsvTalkerId=0, version=0)>",
        )

    def testMgaDbd(self):
        res = UBXReader.parse(self.mga_dbd)
        self.assertEqual(
            str(res),
            "<UBX(MGA-DBD, reserved1=3727165692135864801209549313, data_01=1, data_02=2)>",
        )

    def testMgaFlashAck(self):
        res = UBXReader.parse(self.mga_flash_ack)
        self.assertEqual(
            str(res),
            "<UBX(MGA-FLASH-ACK, type=3, version=1, ack=2, reserved1=0, sequence=1024)>",
        )

    def testCFGVALGET(self):  # test parser of CFG-VALGET CFG-UART1-BAUDRATE
        res = UBXReader.parse(self.cfg_valget)
        self.assertEqual(
            str(res),
            "<UBX(CFG-VALGET, version=0, layer=0, position=0, CFG_UART1_BAUDRATE=9600)>",
        )

    def testCFGVALGET2(self):  # test parse of CFG-VALGET CFG-I2C-ADDRESS
        res = UBXReader.parse(self.cfg_valget2)
        self.assertEqual(
            str(res),
            "<UBX(CFG-VALGET, version=0, layer=0, position=0, CFG_I2C_ADDRESS=85)>",
        )

    def testCFGVALGET3(
        self,
    ):  # test parse of CFG-VALGET CFG-I2C-ADDRESS, CFG-UART1-BAUDRATE, CFG-RATE-NAV
        res = UBXReader.parse(self.cfg_valget3)
        self.assertEqual(
            str(res),
            "<UBX(CFG-VALGET, version=0, layer=0, position=0, CFG_I2C_ADDRESS=85, CFG_UART1_BAUDRATE=9600, CFG_RATE_NAV=35)>",
        )

    def testCFGVALGET4(self):  # test parser of CFG-VALGET CFG-NAVSPG-USRDAT_ROTY
        res = UBXReader.parse(self.cfg_valget4)
        self.assertAlmostEqual(res.CFG_NAVSPG_USRDAT_ROTY, 1.23, 2)

    def testRXMPMPV0(self):  # test parser of RXM-PMP v0 message
        rxm_pmpv0 = b"\xb5b\x02\x72\x0e\x02\x00\x00\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x01\x01"
        for i in range(504):
            n = i % 256
            rxm_pmpv0 += n.to_bytes(1, "little", signed=False)
        rxm_pmpv0 += b"\x01\x02\x01\x00\xcf\xa9"
        res = UBXReader.parse(rxm_pmpv0, validate=VALCKSUM, msgmode=SET)
        self.assertEqual(
            str(res),
            "<UBX(RXM-PMP, version=0, reserved0=131328, timeTag=33620995, uniqueWord1=33620995, uniqueWord2=33620995, serviceIdentifier=257, spare=0, uniqueWordBitErrors=1, userData_01=2, userData_02=3, userData_03=4, userData_04=5, userData_05=6, userData_06=7, userData_07=8, userData_08=9, userData_09=10, userData_10=11, userData_11=12, userData_12=13, userData_13=14, userData_14=15, userData_15=16, userData_16=17, userData_17=18, userData_18=19, userData_19=20, userData_20=21, userData_21=22, userData_22=23, userData_23=24, userData_24=25, userData_25=26, userData_26=27, userData_27=28, userData_28=29, userData_29=30, userData_30=31, userData_31=32, userData_32=33, userData_33=34, userData_34=35, userData_35=36, userData_36=37, userData_37=38, userData_38=39, userData_39=40, userData_40=41, userData_41=42, userData_42=43, userData_43=44, userData_44=45, userData_45=46, userData_46=47, userData_47=48, userData_48=49, userData_49=50, userData_50=51, userData_51=52, userData_52=53, userData_53=54, userData_54=55, userData_55=56, userData_56=57, userData_57=58, userData_58=59, userData_59=60, userData_60=61, userData_61=62, userData_62=63, userData_63=64, userData_64=65, userData_65=66, userData_66=67, userData_67=68, userData_68=69, userData_69=70, userData_70=71, userData_71=72, userData_72=73, userData_73=74, userData_74=75, userData_75=76, userData_76=77, userData_77=78, userData_78=79, userData_79=80, userData_80=81, userData_81=82, userData_82=83, userData_83=84, userData_84=85, userData_85=86, userData_86=87, userData_87=88, userData_88=89, userData_89=90, userData_90=91, userData_91=92, userData_92=93, userData_93=94, userData_94=95, userData_95=96, userData_96=97, userData_97=98, userData_98=99, userData_99=100, userData_100=101, userData_101=102, userData_102=103, userData_103=104, userData_104=105, userData_105=106, userData_106=107, userData_107=108, userData_108=109, userData_109=110, userData_110=111, userData_111=112, userData_112=113, userData_113=114, userData_114=115, userData_115=116, userData_116=117, userData_117=118, userData_118=119, userData_119=120, userData_120=121, userData_121=122, userData_122=123, userData_123=124, userData_124=125, userData_125=126, userData_126=127, userData_127=128, userData_128=129, userData_129=130, userData_130=131, userData_131=132, userData_132=133, userData_133=134, userData_134=135, userData_135=136, userData_136=137, userData_137=138, userData_138=139, userData_139=140, userData_140=141, userData_141=142, userData_142=143, userData_143=144, userData_144=145, userData_145=146, userData_146=147, userData_147=148, userData_148=149, userData_149=150, userData_150=151, userData_151=152, userData_152=153, userData_153=154, userData_154=155, userData_155=156, userData_156=157, userData_157=158, userData_158=159, userData_159=160, userData_160=161, userData_161=162, userData_162=163, userData_163=164, userData_164=165, userData_165=166, userData_166=167, userData_167=168, userData_168=169, userData_169=170, userData_170=171, userData_171=172, userData_172=173, userData_173=174, userData_174=175, userData_175=176, userData_176=177, userData_177=178, userData_178=179, userData_179=180, userData_180=181, userData_181=182, userData_182=183, userData_183=184, userData_184=185, userData_185=186, userData_186=187, userData_187=188, userData_188=189, userData_189=190, userData_190=191, userData_191=192, userData_192=193, userData_193=194, userData_194=195, userData_195=196, userData_196=197, userData_197=198, userData_198=199, userData_199=200, userData_200=201, userData_201=202, userData_202=203, userData_203=204, userData_204=205, userData_205=206, userData_206=207, userData_207=208, userData_208=209, userData_209=210, userData_210=211, userData_211=212, userData_212=213, userData_213=214, userData_214=215, userData_215=216, userData_216=217, userData_217=218, userData_218=219, userData_219=220, userData_220=221, userData_221=222, userData_222=223, userData_223=224, userData_224=225, userData_225=226, userData_226=227, userData_227=228, userData_228=229, userData_229=230, userData_230=231, userData_231=232, userData_232=233, userData_233=234, userData_234=235, userData_235=236, userData_236=237, userData_237=238, userData_238=239, userData_239=240, userData_240=241, userData_241=242, userData_242=243, userData_243=244, userData_244=245, userData_245=246, userData_246=247, userData_247=248, userData_248=249, userData_249=250, userData_250=251, userData_251=252, userData_252=253, userData_253=254, userData_254=255, userData_255=0, userData_256=1, userData_257=2, userData_258=3, userData_259=4, userData_260=5, userData_261=6, userData_262=7, userData_263=8, userData_264=9, userData_265=10, userData_266=11, userData_267=12, userData_268=13, userData_269=14, userData_270=15, userData_271=16, userData_272=17, userData_273=18, userData_274=19, userData_275=20, userData_276=21, userData_277=22, userData_278=23, userData_279=24, userData_280=25, userData_281=26, userData_282=27, userData_283=28, userData_284=29, userData_285=30, userData_286=31, userData_287=32, userData_288=33, userData_289=34, userData_290=35, userData_291=36, userData_292=37, userData_293=38, userData_294=39, userData_295=40, userData_296=41, userData_297=42, userData_298=43, userData_299=44, userData_300=45, userData_301=46, userData_302=47, userData_303=48, userData_304=49, userData_305=50, userData_306=51, userData_307=52, userData_308=53, userData_309=54, userData_310=55, userData_311=56, userData_312=57, userData_313=58, userData_314=59, userData_315=60, userData_316=61, userData_317=62, userData_318=63, userData_319=64, userData_320=65, userData_321=66, userData_322=67, userData_323=68, userData_324=69, userData_325=70, userData_326=71, userData_327=72, userData_328=73, userData_329=74, userData_330=75, userData_331=76, userData_332=77, userData_333=78, userData_334=79, userData_335=80, userData_336=81, userData_337=82, userData_338=83, userData_339=84, userData_340=85, userData_341=86, userData_342=87, userData_343=88, userData_344=89, userData_345=90, userData_346=91, userData_347=92, userData_348=93, userData_349=94, userData_350=95, userData_351=96, userData_352=97, userData_353=98, userData_354=99, userData_355=100, userData_356=101, userData_357=102, userData_358=103, userData_359=104, userData_360=105, userData_361=106, userData_362=107, userData_363=108, userData_364=109, userData_365=110, userData_366=111, userData_367=112, userData_368=113, userData_369=114, userData_370=115, userData_371=116, userData_372=117, userData_373=118, userData_374=119, userData_375=120, userData_376=121, userData_377=122, userData_378=123, userData_379=124, userData_380=125, userData_381=126, userData_382=127, userData_383=128, userData_384=129, userData_385=130, userData_386=131, userData_387=132, userData_388=133, userData_389=134, userData_390=135, userData_391=136, userData_392=137, userData_393=138, userData_394=139, userData_395=140, userData_396=141, userData_397=142, userData_398=143, userData_399=144, userData_400=145, userData_401=146, userData_402=147, userData_403=148, userData_404=149, userData_405=150, userData_406=151, userData_407=152, userData_408=153, userData_409=154, userData_410=155, userData_411=156, userData_412=157, userData_413=158, userData_414=159, userData_415=160, userData_416=161, userData_417=162, userData_418=163, userData_419=164, userData_420=165, userData_421=166, userData_422=167, userData_423=168, userData_424=169, userData_425=170, userData_426=171, userData_427=172, userData_428=173, userData_429=174, userData_430=175, userData_431=176, userData_432=177, userData_433=178, userData_434=179, userData_435=180, userData_436=181, userData_437=182, userData_438=183, userData_439=184, userData_440=185, userData_441=186, userData_442=187, userData_443=188, userData_444=189, userData_445=190, userData_446=191, userData_447=192, userData_448=193, userData_449=194, userData_450=195, userData_451=196, userData_452=197, userData_453=198, userData_454=199, userData_455=200, userData_456=201, userData_457=202, userData_458=203, userData_459=204, userData_460=205, userData_461=206, userData_462=207, userData_463=208, userData_464=209, userData_465=210, userData_466=211, userData_467=212, userData_468=213, userData_469=214, userData_470=215, userData_471=216, userData_472=217, userData_473=218, userData_474=219, userData_475=220, userData_476=221, userData_477=222, userData_478=223, userData_479=224, userData_480=225, userData_481=226, userData_482=227, userData_483=228, userData_484=229, userData_485=230, userData_486=231, userData_487=232, userData_488=233, userData_489=234, userData_490=235, userData_491=236, userData_492=237, userData_493=238, userData_494=239, userData_495=240, userData_496=241, userData_497=242, userData_498=243, userData_499=244, userData_500=245, userData_501=246, userData_502=247, userData_503=1, userData_504=2, fecBits=1, ebno=0.0, reserved1=0)>",
        )

    def testRXMPMPV1(self):  # test parser of RXM-PMP v1 message
        rxm_pmpv1 = b"\xb5b\x02\x72\x23\x00\x01\x00\x0b\x00\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x00\x01\x01\x02\x01\x00"
        for i in range(11):
            rxm_pmpv1 += i.to_bytes(1, "little", signed=False)
        rxm_pmpv1 += b"\x00\x20"
        res = UBXReader.parse(rxm_pmpv1, msgmode=SET)
        self.assertEqual(
            str(res),
            "<UBX(RXM-PMP, version=1, reserved0=0, numBytesUserData=11, timeTag=67305985, uniqueWord1=67305985, uniqueWord2=67305985, serviceIdentifier=513, spare=0, uniqueWordBitErrors=1, fecBits=513, ebno=0.125, reserved1=0, userData_01=0, userData_02=1, userData_03=2, userData_04=3, userData_05=4, userData_06=5, userData_07=6, userData_08=7, userData_09=8, userData_10=9, userData_11=10)>",
        )

    def testRXMRLMS(self):  # test parser of RXM-RLM-S message
        rxm_rlms = b"\xb5b\x02\x59\x10\x00\x00\x01\x00\x00"
        for i in range(8):
            rxm_rlms += i.to_bytes(1, "little", signed=False)
        rxm_rlms += b"\x00\x01\x02\x00\x8b\xbd"
        res = UBXReader.parse(rxm_rlms)
        self.assertEqual(
            str(res),
            "<UBX(RXM-RLM, version=0, type=1, svId=0, reserved0=0, beacon=506097522914230528, message=0, params=513, reserved1=0)>",
        )

    def testRXMRLML(self):  # test parser of RXM-RLM-L message
        rxm_rlms = b"\xb5b\x02\x59\x1c\x00\x00\x02\x00\x00"
        for i in range(8):
            rxm_rlms += i.to_bytes(1, "little", signed=False)
        rxm_rlms += b"\x00"
        for i in range(12):
            rxm_rlms += i.to_bytes(1, "little", signed=False)
        rxm_rlms += b"\x00\x01\x02\xda\x81"
        res = UBXReader.parse(rxm_rlms)
        self.assertEqual(
            str(res),
            "<UBX(RXM-RLM, version=0, type=2, svId=0, reserved0=0, beacon=506097522914230528, message=0, params=3416467015609337987117220096, reserved1=131328)>",
        )

    def testMONSPAN2(self):  # test parser of MON-SPAN message (repeating groups empty)
        mon_span = b"\xb5b\x0a\x31\x04\x00\x00\x00\x01\x02"
        mon_span += b"\x42\xc3"
        res = UBXReader.parse(mon_span)
        self.assertEqual(
            str(res), "<UBX(MON-SPAN, version=0, numRfBlocks=0, reserved0=513)>"
        )

    def testMGAINI1(self):  # test parser of MGA-INI input message with kwargs
        res = UBXReader.parse(self.mga_ini1, msgmode=SET)
        self.assertEqual(
            str(res),
            "<UBX(MGA-INI-POS-LLH, type=1, version=0, reserved0=513, lat=6.7305985, lon=6.7305985, alt=67305985, posAcc=67305985)>",
        )

    def testMGAINI2(self):  # test parser of MGA-INI input message with args
        res = UBXReader.parse(self.mga_ini1, validate=VALCKSUM, msgmode=SET)
        self.assertEqual(
            str(res),
            "<UBX(MGA-INI-POS-LLH, type=1, version=0, reserved0=513, lat=6.7305985, lon=6.7305985, alt=67305985, posAcc=67305985)>",
        )

    def testSETPOLL(self):  # test auto detection of SET or POLL mode
        msg = UBXMessage.config_poll(0,0,["CFG_UART1_BAUDRATE", 0x40530001]).serialize()
        res = UBXReader.parse(msg, msgmode=SETPOLL)
        self.assertEqual(
            str(res),
            "<UBX(CFG-VALGET, version=0, layer=0, position=0, keys_01=1079115777, keys_02=1079181313)>",
        )
        msg = UBXMessage.config_set(0,0,[("CFG_UART1_BAUDRATE", 9600), (0x40530001, 38400)]).serialize()
        res = UBXReader.parse(msg, msgmode=SETPOLL)
        self.assertEqual(
            str(res),
            "<UBX(CFG-VALSET, version=0, ram=0, bbr=0, flash=0, action=0, reserved0=0, CFG_UART1_BAUDRATE=9600, CFG_UART2_BAUDRATE=38400)>",
        )

    def testESFSTATUS(self):  # test parser of ESF-STATUS message
        res = UBXReader.parse(self.esf_status)
        # print(res)
        self.assertEqual(
            str(res),
            #"<UBX(ESF-STATUS, iTOW=23:59:42, version=2, reserved0=548, fusionMode=0, reserved1=0, numSens=6, type_01=5, used_01=0, ready_01=1, calibStatus_01=0, timeStatus_01=1, freq_01=0, badMeas_01=0, badTTag_01=0, missingMeas_01=0, noisyMeas_01=0, type_02=13, used_02=0, ready_02=0, calibStatus_02=0, timeStatus_02=1, freq_02=10, badMeas_02=0, badTTag_02=0, missingMeas_02=0, noisyMeas_02=0, type_03=14, used_03=0, ready_03=1, calibStatus_03=0, timeStatus_03=1, freq_03=10, badMeas_03=0, badTTag_03=0, missingMeas_03=0, noisyMeas_03=0, type_04=0, used_04=0, ready_04=0, calibStatus_04=0, timeStatus_04=1, freq_04=10, badMeas_04=0, badTTag_04=0, missingMeas_04=0, noisyMeas_04=0, type_05=1, used_05=0, ready_05=0, calibStatus_05=0, timeStatus_05=0, freq_05=10, badMeas_05=0, badTTag_05=0, missingMeas_05=0, noisyMeas_05=0, type_06=18, used_06=0, ready_06=1, calibStatus_06=0, timeStatus_06=1, freq_06=0, badMeas_06=0, badTTag_06=0, missingMeas_06=0, noisyMeas_06=0)>",
            "<UBX(ESF-STATUS, iTOW=23:59:42, version=2, wtInitStatus=0, mntAlgStatus=1, insInitStatus=1, imuInitStatus=2, reserved0=0, fusionMode=0, reserved1=0, numSens=6, type_01=5, used_01=0, ready_01=1, calibStatus_01=0, timeStatus_01=1, freq_01=0, badMeas_01=0, badTTag_01=0, missingMeas_01=0, noisyMeas_01=0, type_02=13, used_02=0, ready_02=0, calibStatus_02=0, timeStatus_02=1, freq_02=10, badMeas_02=0, badTTag_02=0, missingMeas_02=0, noisyMeas_02=0, type_03=14, used_03=0, ready_03=1, calibStatus_03=0, timeStatus_03=1, freq_03=10, badMeas_03=0, badTTag_03=0, missingMeas_03=0, noisyMeas_03=0, type_04=0, used_04=0, ready_04=0, calibStatus_04=0, timeStatus_04=1, freq_04=10, badMeas_04=0, badTTag_04=0, missingMeas_04=0, noisyMeas_04=0, type_05=1, used_05=0, ready_05=0, calibStatus_05=0, timeStatus_05=0, freq_05=10, badMeas_05=0, badTTag_05=0, missingMeas_05=0, noisyMeas_05=0, type_06=18, used_06=0, ready_06=1, calibStatus_06=0, timeStatus_06=1, freq_06=0, badMeas_06=0, badTTag_06=0, missingMeas_06=0, noisyMeas_06=0)>",
        )


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()
